import base64
import os
from typing import Literal

from dotenv import load_dotenv
from loguru import logger

logger.print = logger.info

load_dotenv(dotenv_path=".env", override=bool(os.environ.get("ENVIRONMENT")), verbose=True)

# TODO: re-order the constants

os.environ["GITHUB_APP_PEM"] = os.environ.get("GITHUB_APP_PEM") or base64.b64decode(
    os.environ.get("GITHUB_APP_PEM_BASE64", "")
).decode("utf-8")

if os.environ["GITHUB_APP_PEM"]:
    os.environ["GITHUB_APP_ID"] = (
        (os.environ.get("GITHUB_APP_ID") or os.environ.get("APP_ID")).replace("\\n", "\n").strip('"')
    )

TEST_BOT_NAME = "sweep-nightly[bot]"
ENV: Literal["dev", "prod"] = os.environ.get("ENV", "prod")  # dev or prod
DEV: bool = ENV.lower() == "dev"
PROD: bool = ENV.lower() == "prod"
CLI: bool = os.environ.get("CLI", "false").lower() == "true"

BOT_TOKEN_NAME = "bot-token"

SWEEP_HEALTH_URL = os.environ.get("SWEEP_HEALTH_URL")

# For frontend
GITHUB_ID = os.environ.get("GITHUB_ID")
GITHUB_SECRET = os.environ.get("GITHUB_SECRET")

GITHUB_APP_ID = os.environ.get("GITHUB_APP_ID", os.environ.get("APP_ID"))
GITHUB_BOT_USERNAME = os.environ.get("GITHUB_BOT_USERNAME")

GITHUB_LABEL_NAME = os.environ.get("GITHUB_LABEL_NAME", "sweep")
GITHUB_LABEL_COLOR = os.environ.get("GITHUB_LABEL_COLOR", "9400D3")
GITHUB_LABEL_DESCRIPTION = os.environ.get("GITHUB_LABEL_DESCRIPTION", "Sweep your software chores")
GITHUB_APP_PEM = os.environ.get("GITHUB_APP_PEM")
GITHUB_APP_PEM = GITHUB_APP_PEM or os.environ.get("PRIVATE_KEY")
if GITHUB_APP_PEM is not None:
    GITHUB_APP_PEM = GITHUB_APP_PEM.strip(' \n"')  # Remove whitespace and quotes
    GITHUB_APP_PEM = GITHUB_APP_PEM.replace("\\n", "\n")

GITHUB_CONFIG_BRANCH = os.environ.get("GITHUB_CONFIG_BRANCH", "sweep/add-sweep-config")
GITHUB_DEFAULT_CONFIG = os.environ.get(
    "GITHUB_DEFAULT_CONFIG",
    """# Sweep AI turns bugs & feature requests into code changes (https://sweep.dev)
# For details on our config file, check out our docs at https://docs.sweep.dev/usage/config

# This setting contains a list of rules that Sweep will check for. If any of these rules are broken in a new commit, Sweep will create an pull request to fix the broken rule.
rules:
{additional_rules}

# This is the branch that Sweep will develop from and make pull requests to. Most people use 'main' or 'master' but some users also use 'dev' or 'staging'.
branch: 'main'

# By default Sweep will read the logs and outputs from your existing Github Actions. To disable this, set this to false.
gha_enabled: True

# This is the description of your project. It will be used by sweep when creating PRs. You can tell Sweep what's unique about your project, what frameworks you use, or anything else you want.
#
# Example:
#
# description: sweepai/sweep is a python project. The main api endpoints are in sweepai/api.py. Write code that adheres to PEP8.
description: ''

# This sets whether to create pull requests as drafts. If this is set to True, then all pull requests will be created as drafts and GitHub Actions will not be triggered.
draft: False

# This is a list of directories that Sweep will not be able to edit.
blocked_dirs: []
""",
)


ORG_ID = os.environ.get("ORG_ID", None)
POSTHOG_API_KEY = os.environ.get("POSTHOG_API_KEY", "phc_CnzwIB0W548wN4wEGeRuxXqidOlEUH2AcyV2sKTku8n")

if POSTHOG_API_KEY.lower() == "none":
    POSTHOG_API_KEY = None

WHITELISTED_REPOS = os.environ.get("WHITELISTED_REPOS", "").split(",")
BLACKLISTED_USERS = os.environ.get("BLACKLISTED_USERS", "").split(",")

# Default OpenAI
OPENAI_API_KEY = os.environ.get("OPENAI_API_KEY", None)  # this may be none, and it will use azure

OPENAI_API_TYPE = os.environ.get("OPENAI_API_TYPE", "anthropic")
assert OPENAI_API_TYPE in ["anthropic", "azure", "openai"], "Invalid OPENAI_API_TYPE"

AZURE_OPENAI_API_KEY = os.environ.get("AZURE_OPENAI_API_KEY", None)
AZURE_OPENAI_API_VERSION = os.environ.get("AZURE_OPENAI_API_VERSION", None)
AZURE_OPENAI_ENDPOINT = os.environ.get("AZURE_OPENAI_ENDPOINT", None)

AZURE_API_KEY = os.environ.get("AZURE_API_KEY", None)
OPENAI_API_BASE = os.environ.get("OPENAI_API_BASE", None)
OPENAI_API_VERSION = os.environ.get("OPENAI_API_VERSION", None)
AZURE_OPENAI_DEPLOYMENT = os.environ.get("AZURE_OPENAI_DEPLOYMENT", None)

OPENAI_EMBEDDINGS_API_TYPE = os.environ.get("OPENAI_EMBEDDINGS_API_TYPE", "openai")
OPENAI_EMBEDDINGS_AZURE_ENDPOINT = os.environ.get("OPENAI_EMBEDDINGS_AZURE_ENDPOINT", None)
OPENAI_EMBEDDINGS_AZURE_DEPLOYMENT = os.environ.get("OPENAI_EMBEDDINGS_AZURE_DEPLOYMENT", None)
OPENAI_EMBEDDINGS_AZURE_API_VERSION = os.environ.get("OPENAI_EMBEDDINGS_AZURE_API_VERSION", None)

OPENAI_API_ENGINE_GPT35 = os.environ.get("OPENAI_API_ENGINE_GPT35", None)
OPENAI_API_ENGINE_GPT4 = os.environ.get("OPENAI_API_ENGINE_GPT4", None)
MULTI_REGION_CONFIG = os.environ.get("MULTI_REGION_CONFIG", None)
if isinstance(MULTI_REGION_CONFIG, str):
    MULTI_REGION_CONFIG = MULTI_REGION_CONFIG.strip("'").replace("\\n", "\n")
    MULTI_REGION_CONFIG = [item.split(",") for item in MULTI_REGION_CONFIG.split("\n")]

DEFAULT_GPT4_MODEL = os.environ.get("DEFAULT_GPT4_MODEL", "gpt-4o")  # this may break in different envs

FILE_CACHE_DISABLED = os.environ.get("FILE_CACHE_DISABLED", "true").lower() == "true"
GITHUB_BASE_URL = os.environ.get("GITHUB_BASE_URL", "github.com")
GITHUB_ENTERPRISE = GITHUB_BASE_URL != "github.com"
GITHUB_API_BASE_URL = f"https://{GITHUB_BASE_URL}/api/v3" if GITHUB_ENTERPRISE else "https://api.github.com"
GITHUB_GRAPHQL_BASE_URL = f"https://{GITHUB_BASE_URL}/api/graphql" if GITHUB_ENTERPRISE else "https://api.github.com/graphql"

GITHUB_APP_BASE_URL = f"https://{GITHUB_BASE_URL}/github-apps/" if GITHUB_ENTERPRISE else "https://github.com/apps/"
DEFAULT_INSTALL_URL = GITHUB_APP_BASE_URL + (GITHUB_BOT_USERNAME or "").removesuffix("[bot]") + "/installations/select_target"
INSTALL_URL = os.environ.get("INSTALL_URL", DEFAULT_INSTALL_URL)

FRONTEND_URL = (
    os.environ.get("FRONTEND_URL", "")
    or os.environ.get("AUTH_URL", "")
    or os.environ.get("NEXT_AUTH_URL", "")
    or "http://localhost:3000"
).strip("/")

DISABLED_REPOS = os.environ.get("DISABLED_REPOS", "").split(",")

SWEEP_ISSUE_ALLOWLIST: list[str] = os.environ.get("SWEEP_ISSUE_ALLOWLIST", "*").split(",")
FIX_CI_ALLOWLIST: list[str] = os.environ.get("FIX_CI_ALLOWLIST", "*").split(",")

INSTALLATION_ID = os.environ.get("INSTALLATION_ID", None)

AWS_ACCESS_KEY = os.environ.get("AWS_ACCESS_KEY")
AWS_SECRET_KEY = os.environ.get("AWS_SECRET_KEY")
AWS_REGION = os.environ.get("AWS_REGION")
AWS_ANTHROPIC_BASE_URL = os.environ.get("AWS_ANTHROPIC_BASE_URL")
AWS_ANTHROPIC_AVAILABLE = AWS_ACCESS_KEY and AWS_SECRET_KEY and (AWS_REGION or AWS_ANTHROPIC_BASE_URL)

TOGETHER_API_KEY = os.environ.get("TOGETHER_API_KEY", None)

ANTHROPIC_API_KEY = os.environ.get("ANTHROPIC_API_KEY", None)

COHERE_API_KEY = os.environ.get("COHERE_API_KEY", None)

VOYAGE_API_KEY = os.environ.get("VOYAGE_API_KEY", None)

VOYAGE_API_AWS_ACCESS_KEY = os.environ.get("VOYAGE_API_AWS_ACCESS_KEY_ID")
VOYAGE_API_AWS_SECRET_KEY = os.environ.get("VOYAGE_API_AWS_SECRET_KEY")
VOYAGE_API_AWS_REGION = os.environ.get("VOYAGE_API_AWS_REGION")
VOYAGE_API_AWS_ENDPOINT_NAME = os.environ.get("VOYAGE_API_AWS_ENDPOINT_NAME", "voyage-code-2")

VOYAGE_API_USE_AWS = VOYAGE_API_AWS_ACCESS_KEY and VOYAGE_API_AWS_SECRET_KEY and VOYAGE_API_AWS_REGION

# TODO: we need to make this dynamic + backoff
BATCH_SIZE = int(
    os.environ.get(
        "BATCH_SIZE", 64 if VOYAGE_API_KEY else 256
    )  # Voyage only allows 128 items per batch and 120000 tokens per batch
)

JIRA_USER_NAME = os.environ.get("JIRA_USER_NAME", None)
JIRA_API_TOKEN = os.environ.get("JIRA_API_TOKEN", None)
JIRA_URL = os.environ.get("JIRA_URL", None)

SLACK_API_KEY = os.environ.get("SLACK_API_KEY", None)

LICENSE_KEY = os.environ.get("LICENSE_KEY", None)
NO_LICENSE = os.environ.get("NO_LICENSE", "false").lower() == "true"
ALTERNATE_AWS = os.environ.get("ALTERNATE_AWS", "none").lower() == "true"

WEBHOOK_SECRET = os.environ.get("WEBHOOK_SECRET", None)

SENTRY_URL = os.environ.get("SENTRY_URL", None)

CACHE_DIRECTORY = os.environ.get("CACHE_DIRECTORY", "/mnt/caches")

DOCUMENTATION_URL = os.environ.get("DOCUMENTATION_URL", "https://docs.sweep.dev")

assert OPENAI_API_KEY or AZURE_API_KEY or AZURE_OPENAI_API_KEY or CLI, "OPENAI_API_KEY is required."

CIRCLE_CI_PAT = os.environ.get(
    "CIRCLE_CI_PAT", None
)  # if this is present, we will poll from and get logs from circleci
JENKINS_AUTH = os.environ.get("JENKINS_AUTH", None)

if JENKINS_AUTH:
    JENKINS_AUTH = [tuple(auth.split(":")) for auth in JENKINS_AUTH.split(",") if auth]
    assert all(
        len(auth) == 2 for auth in JENKINS_AUTH
    ), "JENKINS_AUTH must be a comma-separated list of username:api_key pairs"

WHITELISTED_USERS = [user for user in os.environ.get("WHITELISTED_USERS", "").split(",") if user]
assert isinstance(WHITELISTED_USERS, list), "WHITELISTED_USERS must be a list"

DEEPSEEK_API_KEY = os.environ.get("DEEPSEEK_API_KEY")
MISTRAL_API_KEY = os.environ.get("MISTRAL_API_KEY")

PREWARM_REPOS = [
    repo.strip() for repo in os.environ.get("PREWARM_REPOS", "").split(",") if repo.strip()
]  # format: org/repo:branch,org/repo2:branch2

CURL_CA_BUNDLE = os.environ.get("CURL_CA_BUNDLE")
AUTHORIZED_USERNAMES = os.environ.get("AUTHORIZED_USERNAMES", "kevinlu1248,wwzeng1,sweep-support").split(",")
DOCKER_BUILDKIT = os.environ.get("DOCKER_BUILDKIT", "1")



# from rich import Console
import datetime
import os
import pickle
import threading
import time
from itertools import chain, islice

from github import Github
from github.Event import Event
from github.IssueEvent import IssueEvent
from github.Repository import Repository
from rich import print

from sweepai.app import handle_request
from sweepai.config.server import GITHUB_API_BASE_URL
from sweepai.o11y.event_logger import logger

DEBUG = os.environ.get("DEBUG", False)
RECORD_EVENTS = os.environ.get("RECORD_EVENTS", False)
MAX_EVENTS = 30
g = Github(os.environ["GITHUB_PAT"], base_url=GITHUB_API_BASE_URL)
repo_name = os.environ["REPO"]
repo = g.get_repo(repo_name)
if DEBUG:
    logger.debug("Debug mode enabled")


def pascal_to_snake(name):
    return "".join(["_" + i.lower() if i.isupper() else i for i in name]).lstrip("_")


def get_event_type(event: Event | IssueEvent):
    if isinstance(event, IssueEvent):
        return "issues"
    else:
        return pascal_to_snake(event.type)[: -len("_event")]


def stream_events(repo: Repository, timeout: int = 2, offset: int = 2 * 60):
    processed_event_ids = set()
    current_time = time.time() - offset
    current_time = datetime.datetime.fromtimestamp(current_time)
    local_tz = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

    while True:
        events_iterator = chain(
            islice(repo.get_events(), MAX_EVENTS),
            islice(repo.get_issues_events(), MAX_EVENTS),
        )
        for i, event in enumerate(events_iterator):
            if event.id not in processed_event_ids:
                local_time = event.created_at.replace(tzinfo=datetime.timezone.utc).astimezone(local_tz)

                if local_time.timestamp() > current_time.timestamp():
                    yield event
                else:
                    if DEBUG:
                        logger.debug(
                            f"Skipping event {event.id} because it is in the past (local_time={local_time}, current_time={current_time}, i={i})"
                        )
            if DEBUG:
                logger.debug(f"Skipping event {event.id} because it is already handled")
            processed_event_ids.add(event.id)
        time.sleep(timeout)


def handle_event(event: Event | IssueEvent, do_async: bool = True):
    if isinstance(event, IssueEvent):
        payload = event.raw_data
        payload["action"] = payload["event"]
    else:
        payload = {**event.raw_data, **event.payload}
    payload["sender"] = payload.get("sender", payload["actor"])
    payload["sender"]["type"] = "User"
    payload["pusher"] = payload.get("pusher", payload["actor"])
    payload["pusher"]["name"] = payload["pusher"]["login"]
    payload["pusher"]["type"] = "User"
    payload["after"] = payload.get("after", payload.get("head"))
    payload["repository"] = repo.raw_data
    payload["installation"] = {"id": -1}
    logger.info(str(event) + " " + str(event.created_at))
    if RECORD_EVENTS:
        _type = get_event_type(event) if isinstance(event, Event) else "issue"
        pickle.dump(
            event,
            open(
                "tests/events/" + f"{_type}_{payload.get('action')}_{str(event.id)}.pkl",
                "wb",
            ),
        )
    if do_async:
        thread = threading.Thread(target=handle_request, args=(payload, get_event_type(event)))
        thread.start()
        return thread
    else:
        return handle_request(payload, get_event_type(event))


def main():
    print(
        f"\n[bold black on white]  Starting server, listening to events from {repo_name}...  [/bold black on white]\n",
    )
    print(
        f"To create a PR, please create an issue at https://github.com/{repo_name}/issues with a title prefixed with 'Sweep:' or label an existing issue with 'sweep'. The events will be logged here, but there may be a brief delay.\n"
    )
    for event in stream_events(repo):
        handle_event(event)


if __name__ == "__main__":
    main()



import uuid
from dataclasses import dataclass, field
from typing import Literal, Optional

from sweepai.utils.diff import generate_diff


@dataclass
class CodeSuggestion:
    file_path: str
    original_code: str
    new_code: str
    file_contents: str = ""
    id: str = field(default_factory=lambda: "pyid-" + str(uuid.uuid4()))

    def to_camel_case(self):
        return {
            "filePath": self.file_path,
            "originalCode": self.original_code,
            "newCode": self.new_code,
        }
    
    @classmethod
    def from_camel_case(cls, data: dict):
        return cls(
            file_path=data["filePath"],
            original_code=data["originalCode"],
            new_code=data["newCode"],
        )
    
    def __str__(self):
        diff = generate_diff(self.original_code, self.new_code)
        return f"CodeSuggestion(file_path={self.file_path}, diff=```{diff}```)"


@dataclass
class StatefulCodeSuggestion(CodeSuggestion):
    state: Literal["pending", "processing", "done", "error"] = "pending"
    error: Optional[str] = None

    def to_camel_case(self):
        return {
            **super().to_camel_case(),
            "state": self.state,
        }
    
    @classmethod
    def from_camel_case(cls, data: dict):
        suggestion = super().from_camel_case(data)
        suggestion.state = data["state"]
        return suggestion



from typing import Literal

# I don't like enums because they cause a lot of unexpected errors when used incorrectly
# For example, Enum.VALUE is not of type string and comparisons won't work properly
UseCase = Literal["chat", "ticket", "ci"]
use_cases: list[UseCase] = list(UseCase.__args__)



"""
List of common prompts used across the codebase.
"""

fix_files_to_change_prompt = """You proposed a plan. However, your proposed plan has the following errors:

<errors>
{error_message}
</errors>

You must resolve these errors before proceeding. Respond in the following format:

<error_resolutions>
<error_resolution>
Error #0: Summary of the error

You will first think step-by-step about the error, and then either rewrite the instructions with the corrected fix, or drop the task.

<thinking>
Analyze extremely carefully in great detail what went wrong, including the file path and the specific code block that needs to be modified. 
If you have failed to copy code verbatim, describe precisely what is different between the code you provided and the code in the actual file. Reference the exact lines in the diff provided showing the difference between your original code and what is in the file.
</thinking>

Then, let's resolve the errors in your proposed plan. You MUST pick ONE of the following options:
a. If you would like to patch the corresponding task of the plan, create a modify block with an index. The index should be equivalent to the error number of this error_resolution block, so it must be one of the following allowed integers: {allowed_indices}.
b. If the error is a file path, correct the file path. This is preferred if the code does not need to be changed.
c. Otherwise, if you absolutely cannot resolve the error, drop the task. 

You must pick exactly ONE option from the three options presented above. Follow this format:

Option a: To patch an invalid modify:

<modify file="file_path" index="0">
Rewritten instructions to resolve the error. Update the original_code and new_code blocks as required, ensuring that the <original_code> block contains the actual code from the file.

Update <original_code> with the necessary changes:
<original_code>
The corrected code from the file verbatim. Abbreviating, missing indents, paraphrasing and placeholder code are NOT permitted. It is absolutely critical that the indentation is correct and matches the source code EXACTLY.
</original_code>

Update <new_code> block with the necessary changes:
<new_code>
Updated new code, based on the corrections in <original_code>. Ensure all newly introduced indents and comments are propagated here.
</new_code>
</modify>

Option b: To fix an invalid file path, such as a non-existent directory (this is preferred if the code does not need to be changed):

Enter "COPIED_FROM_PREVIOUS_MODIFY" verbatim into the modify block to copy the code from the previous change.
Example:
<modify file="file_path" index="0">
COPIED_FROM_PREVIOUS_MODIFY
</modify>

Option c: Otherwise, if you absolutely cannot resolve the error, drop the task like so:

<drop>Index of the task to drop</drop>
</error_resolution>

[additional <error_resolution> blocks as needed, for the same file or different files]
</error_resolutions>"""

gha_files_to_change_system_prompt = """You are an AI assistant for analyzing failing errors in a developer's code. You will be provided code files, a description of the issue, the error log, relevant parts of the codebase, and the changes he's made.

Your role is to analyze the issue and codebase. Reference specific files, functions, variables and code files in your analysis.

Take these steps:
1. Analyze the issue, codebase and existing changes to understand the problem.

2. List ALL new errors that have appeared in the GitHub Action logs and their corresponding root causes. Identify ALL the entities that need to be updated to resolve these errors. You are diligent and won't miss any errors or entities."""

gha_files_to_change_system_prompt_2 = """You are an AI assistant for writing code to fix failing errors in a developer's code. You will be provided code files, a description of the issue, the error log, relevant parts of the codebase, and the changes he's made. You may only modify edit files to resolve the issue.

Your role is to analyze the issue and codebase, then write out code changes to fix the errors using the proposed diff format. Reference specific files, functions, variables and code files in your plan.
Prioritize using existing code and functions to make efficient and maintainable changes. Ensure your suggestions fully resolve the issue.

Take these steps:
Create a set of code to remove and code to add, including all necessary changes to resolve the issue, in the following format:
Step 1. Reference the original code in <original_code> tags, copying them VERBATIM from the file, with correct indentation and whitespace.
    - Do NOT paraphrase or abbreviate the source code.
    - Placeholder comments like "# existing code" are not permitted.
    - Start with the closest header several lines before the target code and end with the closest end of the block or several lines after the code.
Step 2. Write the new code in <new_code> tags, specifying necessary imports and including relevant type definitions, interfaces, and schemas.
    - BE EXACT as this code will replace the mentioned <original_code>.
Step 3. Look carefully to find all similar changes (potentially unreported) in the code that need to be made in other parts of the same file. Address those as well."""

gha_files_to_change_prompt = """Your job is to analyze the provided issue, error log, relevant parts of the codebase, and changes he's made to understand the requested change.

Follow the following XML format:

# 1. Thinking:
<thinking>
a. Summarize what the original GitHub issue is and asks us to do.

b. List ALL the changes made so far in extreme detail. Be absolutely complete. Follow this format:
    - File path 1:
        - Description of first diff hunk in extreme detail.
        - Description of second diff hunk in extreme detail.
        [additional changes as needed]
    - File path 2:
        - Description of first diff hunk in extreme detail.
        - Description of second diff hunk in extreme detail.
        [additional changes as needed]
    [additional files as needed]
</thinking>

# 2. Reflect:
<reflection>
a. List out all the previous error logs that were solved.
b. List out all previous error logs that are still present or only partially solved.
c. Identify the number of errors, and list out the indices of all the new Github Action error logs that you must now solve and potential root causes and solutions.
- Error 1/n: For each error answer the following questions:
    1. What is the error message? Repeat it verbatim. Is this a repeat error? If so, you may skip the rest of the questions and simply reference the first appearance of this error.
    2. What is the root cause of the error? Explain why.
    3. How can we resolve the error? Be complete and precise.
    4. If we make the fix, what are the potential side effects? Will the fix break other parts of the code? For example, if we change a function signature, will we need to update all the calls to that function?
    5. If you identify potential side effects create - Side Effect #n block and describe the changes that need to be made in order to fix the side effect.
[repeat for all errors]
</reflection>"""

gha_files_to_change_prompt_2 = """Now that you've analyzed the issue and error logs, your job is to write code changes to help resolve the errors in his code while also resolving the GitHub issue.
For each unique error log identified in the previous step, you will need to provide a set of code changes to fix the error.

Guidelines:
<guidelines>
- Always include the full file path and reference the provided files 
- Prioritize using existing code and utility methods to minimize writing new code
- Break the task into small steps, with each <modify> section for each logical code block worth of change. Use multiple <modify> blocks for the same file if there are multiple distinct changes to make in that file, such as for imports.
- A <modify> block must contain exactly one change in one <new_code> tag.
- To remove code, replace it with empty <new_code> tags.
- Do not make a change that has already been made by the developer.
<guidelines>

Please use the following XML format for your response:

<plan>
There are a total of n errors. List ALL the types of error messages in the current error logs and their root causes. Follow this format:

<error_analysis index="1">
Error message 1/n: Identify the error message.
1. Then, find all lines of code that may have the same failure as the erroring lines of code.
2. Identify the root cause of the error, i.e. whether the error is due to a missing change in the tests or the source code. Most of the time, the test case has yet to be updated.
3. Explain how to resolve the error in the test case. Be complete and precise. Remember that to resolve the error in a way such that the test case is still valid. Do not simply apply a band-aid solution to make the error go away.
4. Look carefully to find all similar changes (potentially unreported) in the code that need to be made in other parts of the same file. Address those as well.

Then, based on the analysis, propose a fix by following the format below. If the error has already been fixed, you can skip this step.

<modify file="file_path"> 
Instructions for modifying one section of the file. Each block must have exactly one original_code and one new_code block.

a. Describe the section of code that needs to be modified.
<original_code>
Copy the original_code here VERBATIM from the file.
Do NOT paraphrase or abbreviate the source code.
Placeholder comments like "# existing code" are not permitted.
Start a few lines before the target code to change and end a few lines after.
Do not edit the same area of code more than once to avoid merge conflicts with other modifies.
</original_code>

b. Describe the changes that need to be made to the code.
<new_code>
Write the new code in <new_code> tags, specifying necessary imports and referencing relevant type definitions, interfaces, and schemas. BE EXACT as this code will replace the mentioned <original_code>. This code MUST be different from the original_code.
</new_code>

Use multiple <modify> blocks for the same file to separate distinct changes, such as for imports.
</modify>

Then, determine if there are similar issues that we should also resolve. Make as many additional modify blocks as needed until ALL similar issues are resolved.
Any issue that doesn't have a corresponding modify block will not be fixed, you must make modify blocks for every single issue identified.
Do not attempt to fix multiple issues with a single modify block, each issue must have its own modify block.
</error_analysis>
[additional <error_analysis> blocks as needed, for ALL error messages in the error logs
</plan>"""

plan_selection_prompt = """Critique the pros and cons of each plan based on the following guidelines, prioritizing thoroughness and correctness over potential performance overhead: 
- Correctness: The code change should fully address the original issue or requirement without introducing new bugs, security vulnerabilities, or performance problems. Follow defensive programming practices, such as avoiding implicit assumptions, validating inputs, and handling edge cases. Consider the potential impact on all relevant data structures and ensure the solution maintains data integrity and consistency. Thoroughness is a top priority. 
- Backwards Compatibility: When possible, avoid breaking changes to public APIs, data formats, or behaviors that existing code depends on. 
- Clarity: The code change should be readable, well-structured, and easy for other developers to understand and maintain. Follow existing conventions and style guides, and include documentation and comments for complex or non-obvious logic. 
- Simplicity: Strive for a solution that is as simple as possible while still being complete and correct. Favor straightforward and easily understandable code. Performance overhead should not be a factor in evaluating simplicity. 
- Integration: Assess how well the change fits with the overall architecture and design of the system. Avoid tightly coupling components or introducing new dependencies that could complicate future development or deployment. After evaluating the plans against these criteria, select the one that provides the most thorough and correct solution within the specific context and constraints of the project. Prioritize long-term maintainability and architectural integrity.

Respond using the following XML format:

<final_plan>
[Insert the final plan here, including any modifications or improvements based on the feedback and dialogue. Explain how the plan aligns with the guidelines and why it was chosen over the alternatives.]
</final_plan>

Here is an example response format:

<final_plan>
<modify file="example.py">
[Example instructions here]
</modify>
...
<modify file="anotherexamplefile.py">
[More example instructions here]
</modify>
[Your explanation of why this plan was chosen and how it aligns with the guidelines and any modications made to this plan]
</final_plan>"""



from github import Github

from sweepai.core.github_utils import get_github_client, get_installation_id
from sweepai.web.events import Account, Installation, IssueRequest


def fetch_issue_request(
    org_name: str,
    repo_name: str,
    issue_number: str,
    issue_url: str = "",
    edited: bool = True,
    __version__: str = "0",
):

    print("Fetching installation ID...")
    installation_id = get_installation_id(org_name)
    print("Fetching access token...")
    _token, g = get_github_client(installation_id)
    g: Github = g
    print("Fetching repo...")
    issue = g.get_repo(f"{org_name}/{repo_name}").get_issue(int(issue_number))

    issue_request = IssueRequest(
        action="labeled" if not edited else "edited",
        issue=IssueRequest.Issue(
            title=issue.title,
            number=int(issue_number),
            html_url=issue_url,
            user=IssueRequest.Issue.User(
                login=issue.user.login,
                type="User",
            ),
            body=issue.body,
            labels=[
                IssueRequest.Issue.Label(
                    name="sweep",
                ),
            ],
            assignees=None,
            pull_request=None,
        ),
        repository=IssueRequest.Issue.Repository(
            full_name=issue.repository.full_name,
            description=issue.repository.description,
        ),
        assignee=IssueRequest.Issue.Assignee(login=issue.user.login),
        installation=Installation(
            id=installation_id,
            account=Account(
                id=issue.user.id,
                login=issue.user.login,
                type="User",
            ),
        ),
        sender=IssueRequest.Issue.User(
            login=issue.user.login,
            type="User",
        ),
    )
    return issue_request



from __future__ import annotations

import json
import os
import re
import subprocess
from dataclasses import dataclass, field
from functools import cached_property
from typing import ClassVar, Iterator, Literal, Optional

import joblib
from diskcache import Cache
from loguru import logger
from tqdm import tqdm
from tree_sitter import Node

from sweepai.config.client import SweepConfig
from sweepai.config.server import CACHE_DIRECTORY
from sweepai.modify.validate.code_validators import EXT_TO_LANGUAGE, get_language
from sweepai.search.agent.code_tree import CodeTree
from sweepai.utils.cache import hash_code
from sweepai.utils.timer import Timer

entities_cache = Cache(f"{CACHE_DIRECTORY}/entities_cache")
ENTITIES_CACHE_VERSION = "v0.1"

# Future: migrate to https://github.com/tree-sitter/tree-sitter-graph
# TODO: python built-in methods are not being recognized properly
# TODO: add better support for raw import statements and aliases
# for testing s exp queries, go to https://tree-sitter.github.io/tree-sitter/playground
# all s_exp stored in sweepai/search/agent/tree_sitter_querys/[language].scm


def clip(value: int, min_value: int, max_value: int) -> int:
    return max(min_value, min(value, max_value))


@dataclass
class LanguageConfig:
    language: str
    s_exp_query: str
    builtin_variables: set[str]
    builtin_modules: set[str]
    builtin_methods: set[str]
    config_directory: ClassVar[str] = "sweepai/search/agent/tree_sitter_queries"

    @classmethod
    def from_language(cls, language: str):
        s_exp_query = open(f"{cls.config_directory}/{language}.scm").read()
        metadata_path = f"{cls.config_directory}/{language}.json"
        if os.path.exists(metadata_path):
            with open(metadata_path) as f:
                metadata = json.load(f)
        else:
            metadata = {}
        return cls(
            language=language,
            s_exp_query=s_exp_query,
            builtin_variables=set(metadata.get("variables", [])),
            builtin_modules=set(metadata.get("modules", [])),
            builtin_methods=set(metadata.get("methods", [])),
        )

    @cached_property
    def builtin_modules_pattern(self):
        return re.compile("^(" + f"|".join(self.builtin_modules) + ").*")

    def is_builtin_module(self, module: str) -> bool:
        module = module.strip('"')
        return self.builtin_modules_pattern.fullmatch(module)

    @classmethod
    def get_available_languages(cls):
        return [
            file_path.removesuffix(".scm")
            for file_path in os.listdir(cls.config_directory)
            if file_path.endswith(".scm")
        ]

    @classmethod
    def load_all(cls):
        return {language: cls.from_language(language) for language in cls.get_available_languages()}


LANGUAGE_CONFIGS = LanguageConfig.load_all()


def parse_potential_entity_from_line(query_to_get_entity_from: str):
    if "::" in query_to_get_entity_from:
        split_query_with_potential_entity = query_to_get_entity_from.split("::")
    elif ":" in query_to_get_entity_from:
        split_query_with_potential_entity = query_to_get_entity_from.split(":")
    else:
        return "", query_to_get_entity_from
    file_path, entity_name = (
        split_query_with_potential_entity[0],
        split_query_with_potential_entity[1],
    )
    return file_path, entity_name


# Based on: https://tree-sitter.github.io/tree-sitter/code-navigation-systems#examples, with a few modifications
node_types = [
    "definition.import",
    "definition.constant",
    "definition.class",
    "definition.function",
    "definition.method",
    "definition.parameter",
    "definition.interface",
    "definition.variable",
    "definition.type",
    "reference.call",
    "reference.class",
    "reference.implementation",
    "reference.type",
]


class Serializable:
    """Can use Pydantic but Pydantic is a bit of a pita to work with."""

    def to_dict(self):
        return self.__dict__

    @classmethod
    def from_dict(cls, data):
        return cls(**data)


@dataclass
class Entity(Serializable):
    name: str
    type: Literal[
        "definition.constant",
        "definition.class",
        "definition.function",
        "definition.method",
        "definition.interface",
        "definition.variable",
        "definition.type",
        "reference.call",
        "reference.class",
        "reference.implementation",
        "reference.type",
    ]

    start_line: int
    end_line: int
    start_char: int
    end_char: int

    file_path: str
    language: str
    role: Literal["definition", "reference"]
    contents: str | None = None
    metadata: dict[str, dict] = field(default_factory=dict)
    parent: Optional[Entity] = None

    @classmethod
    def from_node(
        cls,
        name: str,
        type: Literal[
            "definition.constant",
            "definition.class",
            "definition.function",
            "definition.method",
            "definition.interface",
            "definition.variable",
            "definition.type",
            "reference.call",
            "reference.class",
            "reference.implementation",
            "reference.type",
        ],
        node: Node,
        file_path: str,
        parent: Optional[Entity] = None,
        **kwargs,
    ):
        role, entity_type = type.split(".")
        start_char = node.start_byte
        end_char = node.end_byte

        contents = " " * node.start_point[1] + node.text.decode()
        parent_node = node.parent
        if (
            parent_node
            and max((parent_node.end_point[0] - parent_node.start_point[0]), 1)
            / max(node.end_point[0] - node.start_point[0], 1)
            <= 2
        ):
            # continuously check parent
            contents = " " * parent_node.start_point[1] + parent_node.text.decode()
            start_char = parent_node.start_byte
            end_char = parent_node.end_byte
        data = {
            "name": name,
            "type": type,
            "start_line": node.start_point[0],
            "end_line": node.end_point[0],
            "file_path": file_path,
            "start_char": start_char,
            "end_char": end_char,
            "parent": parent,
            "contents": contents,
        }
        if entity_type == "import":
            return Import(**data, **kwargs)
        elif role == "definition":
            return Definition(**data, **kwargs)
        else:
            return Reference(**data, **kwargs)

    def hydrate(self, file_contents: str):
        self.contents = file_contents[self.start_char : self.end_char + 1]
        for attr in ["parent", "object", "referenced_entity", "resolved_entity"]:
            if hasattr(self, attr) and (entity := getattr(self, attr)):
                entity.hydrate(file_contents)

    @property
    def entity_type(self):
        return self.type.split(".")[-1]

    @property
    def file_denotation(self):
        return f"{self.file_path}:{self.start_line}-{self.end_line}"

    @property
    def full_reference(self):
        return f"{self.file_path}:{self.full_name}"

    @property
    def full_name(self):
        if self.parent:
            return f"{self.parent.full_name}.{self.name}"
        return self.name

    def __str__(self):
        return self.full_name

    def get_parents(self):
        if self.parent:
            return self.parent.get_parents() + [self.parent]
        return []

    def to_dict(self):
        data = {**super().to_dict()}
        if "contents" in data:
            del data["contents"]
        if "node" in data:
            del data["node"]
        for attr in ["parent", "object", "referenced_entity", "resolved_entity"]:
            if attr in data and data[attr] and isinstance(data[attr], Serializable):
                data[attr] = data[attr].to_dict()
        return data

    @classmethod
    def from_dict(cls, **kwargs):
        role, entity_type = kwargs["type"].split(".")
        for attr in ["parent", "object", "referenced_entity", "resolved_entity"]:
            if attr in kwargs and kwargs[attr]:
                kwargs[attr] = cls.from_dict(**kwargs[attr])
        if entity_type == "import":
            return Import(
                **kwargs,
            )
        elif role == "definition":
            return Definition(
                **kwargs,
            )
        else:
            return Reference(
                **kwargs,
            )


@dataclass
class Definition(Entity):
    scope: Literal["local", "global"] = "global"
    role: Literal["definition"] = "definition"
    type_annotation: str | None = None

    def __post_init__(self):
        if self.entity_type == "function" and self.parent and self.parent.entity_type == "class":
            self.type = "definition.method"
        if self.parent and self.parent.entity_type != "class":
            self.scope = "local"


@dataclass
class Import(Definition):
    module: str = ""

    def __post_init__(self):
        if self.language == "go":
            self.module = self.name
            self.name = self.name.strip('"').split("/")[-1]
        elif self.language == "python":
            if not self.module.endswith(".py"):
                self.module = self.module.replace(".", "/")
                self.module += ".py"


@dataclass
class Reference(Entity):
    referenced_entity: Optional[Entity] = None  # e.g. the import that contains this reference
    resolved_entity: Optional[Entity] = None  # e.g. the actual definition of this reference
    origin_type: Literal["defined", "builtin", "imported", "unknown"] = "unknown"
    role: Literal["reference"] = "reference"
    object: Entity | None = None

    def __post_init__(self):
        # TODO: classify into variable, types, functions etc.
        builtins = LANGUAGE_CONFIGS[self.language].builtin_variables
        if self.name in builtins:
            self.origin_type = "builtin"
        elif self.name in LANGUAGE_CONFIGS[self.language].builtin_methods and self.object:
            self.origin_type = "builtin"

    @property
    def full_name(self):
        name = self.name
        if self.object:
            name = f"{self.object}.{self.name}"
        if self.parent:
            return f"{self.parent.full_name}.{name}"
        return name


def get_parents(node: Node) -> Iterator[Node]:
    parent = node.parent
    while parent:
        yield parent
        parent = parent.parent


@dataclass
class FileEntities(Serializable):
    """
    TODO: add general query support
    """

    key: str
    definitions: dict[str, Definition] = field(default_factory=dict)
    references: list[Reference] = field(default_factory=list)
    node_id_mapping: dict[int, str] = field(default_factory=dict)

    def all_entities(self) -> list[Entity]:
        return sorted([*self.definitions.values(), *self.references], key=lambda x: x.start_line)

    @classmethod
    def get_query(cls, language_name: str):
        if not hasattr(cls, "query_cache"):
            cls.query_cache = {}
        query = cls.query_cache.get(language_name, None)
        if query is None:
            language_config = LANGUAGE_CONFIGS[language_name]
            language = get_language(language_name)
            query = language.query(language_config.s_exp_query)
            cls.query_cache[language_name] = query
        return query

    @classmethod
    def from_file(cls, file_path: str, cwd: str = "", key: str = "", no_cache: bool = False) -> FileEntities | None:
        language_name = file_path.split(".")[-1]
        language_name = EXT_TO_LANGUAGE.get(language_name, language_name)
        if language_name not in LANGUAGE_CONFIGS:
            return None
        if not os.path.exists(file_path):
            return None  # This edge case only happens when back-testing

        if not no_cache:
            file_contents_for_hash = open(file_path, "r").read()
            key = f"{key}-{hash_code(file_contents_for_hash)}"
            if cached_entities := cls.from_cache(key):
                return cached_entities

        code_tree = CodeTree.from_file(file_path)  # 30% of time spent here
        file_contents = code_tree.code
        file_bytes = code_tree.code.encode()
        query = cls.get_query(language_name)
        matches = query.matches(code_tree.tree.root_node)
        entities_mapping = cls(key=key)
        # entities: dict[str, Entity] = {} # maps names to Entities
        for _id, match in matches:
            match_type = None
            name_node = match["name"]
            node = None
            for key in node_types:
                if key in match:
                    match_type = key
                    node = match[key]
            if match_type:
                name = file_bytes[name_node.start_byte : name_node.end_byte].decode("utf-8")
                parent_name = ""
                kwargs = {}
                for parent in get_parents(node):
                    if parent.id in entities_mapping.node_id_mapping:
                        parent_name = entities_mapping.node_id_mapping[parent.id]
                        break
                if match_type == "definition.import" and "module" in match:
                    kwargs["module"] = file_contents[match["module"].start_byte : match["module"].end_byte]
                if match_type == "definition.variable" and "type" in match:
                    kwargs["type_annotation"] = file_contents[match["type"].start_byte : match["type"].end_byte]
                if "object" in match:
                    if match["object"].id in entities_mapping.node_id_mapping:
                        kwargs["object"] = entities_mapping.definitions[
                            entities_mapping.node_id_mapping[match["object"].id]
                        ]
                    else:
                        object_name = file_contents[match["object"].start_byte : match["object"].end_byte]
                        kwargs["object"] = Entity.from_node(
                            name=object_name,
                            type="reference.variable",
                            node=match["object"],
                            file_path=file_path.removeprefix(cwd + os.sep),
                            parent=entities_mapping.definitions.get(parent_name),
                            language=language_name,
                        )
                entity = Entity.from_node(
                    name,
                    match_type,
                    node,
                    file_path.removeprefix(cwd + os.sep),
                    entities_mapping.definitions.get(parent_name),
                    language=language_name,
                    **kwargs,
                )
                entities_mapping.add_entity(entity, node.id)
        # Single file resolution
        for entity in entities_mapping.references:
            entities_mapping.resolve_entity(entity)
        # entities_mapping.to_cache()
        return entities_mapping

    def get_entity(self, name: str, scope: list[str] = []):
        current_scope = scope
        while True:
            full_name = ".".join([frame for frame in current_scope] + [name])
            if full_name in self.definitions:
                # match up the types
                return self.definitions[full_name]
            if len(current_scope) == 0:
                break
            current_scope.pop()

    def add_entity(self, entity: Entity, node_id: int):
        if entity.role == "definition":
            self.definitions[entity.full_name] = entity
            self.node_id_mapping[node_id] = entity.full_name
        else:
            self.references.append(entity)

    def resolve_entity(self, entity: Reference):
        language_config = LANGUAGE_CONFIGS[entity.language]
        if entity.origin_type != "builtin":
            referenced_entity = self.get_entity(entity.name, [parent.name for parent in entity.get_parents()])
            if referenced_entity:
                entity.referenced_entity = referenced_entity
                entity.origin_type = "defined"
                if referenced_entity.type_annotation:
                    entity.type_annotation = (
                        referenced_entity.type_annotation
                    )  # not sure if this is the best way to do it
                    # use a node later
                    if entity.type_annotation in self.definitions:
                        type_definition_entity = self.definitions[entity.type_annotation]
                        if type_definition_entity.entity_type == "import":
                            if language_config.is_builtin_module(type_definition_entity.module):
                                entity.origin_type = "builtin"
                            else:
                                entity.origin_type = "imported"
                        elif type_definition_entity.entity_type == "type":
                            entity.origin_type = "defined"
                if referenced_entity.entity_type == "import":
                    if language_config.is_builtin_module(referenced_entity.module):
                        entity.origin_type = "builtin"
                    else:
                        entity.origin_type = "imported"
                else:
                    entity.resolved_entity = referenced_entity
            if not entity.referenced_entity:
                if entity.object:
                    # check object in the parent scope
                    referenced_entity = self.get_entity(
                        entity.object.name,
                        [parent.name for parent in entity.get_parents()],
                    )
                    if referenced_entity:
                        entity.referenced_entity = referenced_entity
                        type_name = referenced_entity.type_annotation
                        if type_name:
                            type_entity = self.get_entity(
                                type_name,
                                [parent.name for parent in entity.get_parents()],
                            )
                            if type_entity:
                                referenced_type_entity = self.get_entity(
                                    referenced_entity.type_annotation,
                                    [parent.name for parent in entity.get_parents()],
                                )
                                if isinstance(referenced_type_entity, Import) and language_config.is_builtin_module(
                                    referenced_type_entity.module
                                ):
                                    entity.origin_type = "builtin"
                                else:
                                    entity.origin_type = "imported"
                                return entity
                        referenced_name = (
                            entity.referenced_entity.module
                            if isinstance(entity.referenced_entity, Import)
                            else entity.referenced_entity.name
                        )
                        if language_config.is_builtin_module(referenced_name):
                            entity.origin_type = "builtin"
                        else:
                            entity.origin_type = "imported"
                        return entity
                elif entity.language == "go" or entity.language == "java":
                    entity.origin_type = "imported"
        return entity

    def hydrate(self, file_contents: str):
        for entity in self.definitions.values():
            entity.hydrate(file_contents)
        for entity in self.references:
            entity.hydrate(file_contents)

    @classmethod
    def from_dict(cls, data: dict, key: str):
        try:
            return cls(
                key=key,
                definitions={name: Entity.from_dict(**entity) for name, entity in data["definitions"].items()},
                references=[Entity.from_dict(**entity) for entity in data["references"]],
            )
        except Exception:
            logger.error(f"Error deserializing entity: {data}")
            return None

    def to_dict(self):
        return {
            "key": self.key,
            "definitions": {name: entity.to_dict() for name, entity in self.definitions.items()},
            "references": [entity.to_dict() for entity in self.references],
        }

    def to_cache(self):
        with Timer(f"serializing: {self.key}", min_time=0.01):
            obj = self.to_dict()
        entities_cache[self.key] = obj

    @classmethod
    def from_cache(cls, key: str):
        if results := entities_cache.get(key, None):
            with Timer(f"deserializing: {key}", min_time=0.05):
                return cls.from_dict(results, key=key)


def gitignored_walk(directory: str, sweep_config: SweepConfig) -> Iterator[str]:
    with subprocess.Popen(
        ["git", "ls-files", "--recurse-submodules"],
        stdout=subprocess.PIPE,
        cwd=directory,
    ) as proc:
        for line in proc.stdout:
            file_path = line.decode().strip()
            if not sweep_config.is_file_excluded_aggressive(directory, file_path):  # filters out auto-generated files
                yield file_path


# TODO: handle multi-root workspaces


def serialize(data):
    if isinstance(data, list):
        return [serialize(item) for item in data]
    elif isinstance(data, dict):
        return {key: serialize(value) for key, value in data.items()}
    elif isinstance(data, Serializable):
        return serialize(data.to_dict())
    else:
        return data


def get_file_entities_from_file(file_path: str, dir_path: str, key: str):
    if os.path.isdir(os.path.join(dir_path, file_path)):
        return None
    with Timer(f"FileEntities.from_file({file_path})", min_time=0.1):
        entities = FileEntities.from_file(
            os.path.join(dir_path, file_path),
            cwd=dir_path,
            key=f"{key}:{file_path}",
            no_cache=True,
        )
    return (file_path, entities) if entities else None


@dataclass
class EntitiesIndex(Serializable):
    key: str
    # file path -> entity name -> entity
    entities_mapping: dict[str, FileEntities] = field(default_factory=dict)
    definitions_index: dict[str, list[Entity]] = field(default_factory=dict)
    # (file path, entity name) -> entity[]
    usages_index: dict[str, list[Reference]] = field(default_factory=dict)
    entities_search_cache: dict[tuple[str, str], list[Entity]] = field(default_factory=dict)

    def __repr__(self):
        return f"EntitiesIndex({self.key}, {len(self.entities_mapping)} files, {len(self.definitions_index)} definitions, {len(self.usages_index)} usages)"

    @classmethod
    def from_dir(cls, dir_path: str, key: str = "", no_cache: bool = False) -> EntitiesIndex:
        key = key or dir_path
        if not no_cache and (cached_index := entities_cache.get(key)) is not None:
            return cached_index
        files_to_entities = cls.ingest_dir(dir_path, key=key)
        definitions_index: dict[str, list[Entity]] = {}
        for _, entities_mapping in files_to_entities.items():
            for name, entity in entities_mapping.definitions.items():
                if entity.scope == "global":
                    definitions_index.setdefault(entity.name, []).append(entity)
        results = cls(
            key=key,
            entities_mapping=files_to_entities,
            definitions_index=definitions_index,
        )
        if not no_cache:
            with Timer("hydrating entities"):
                results.hydrate(dir_path)
        with Timer("resolving references"):
            results.resolve_references()
        entities_cache[key] = results
        return results

    @staticmethod
    def ingest_dir(
        dir_path: str,
        dirs_to_ingest: list[str] = [],
        key: str = "",
    ) -> dict[str, FileEntities]:
        # Helper functio to ingest entities from a directory
        files_to_entities = {}
        sweep_config = SweepConfig()
        file_list = list(gitignored_walk(dir_path, sweep_config))
        # TODO: make this configurable
        # TODO: make caching better
        if dirs_to_ingest:
            file_list = [
                file_path for file_path in file_list if any(dir_path in file_path for dir_path in dirs_to_ingest)
            ]
        with Timer("ingesting files"):
            workers = clip(len(file_list), 1, 8)
            # specify to use thread
            try:

                results = joblib.Parallel(n_jobs=workers, backend="multiprocessing")(
                    joblib.delayed(get_file_entities_from_file)(file_path, dir_path, key) for file_path in file_list
                )
            except Exception as e:
                logger.error(f"Error ingesting entities: {e}")
                results = [get_file_entities_from_file(file_path, dir_path, key) for file_path in file_list]
            files_to_entities.update(dict(filter(None, results)))
        return files_to_entities

    def resolve_references(self):
        for entities_mapping in tqdm(
            self.entities_mapping.values(),
            desc="Resolving references",
            total=len(self.entities_mapping),
        ):
            for entity in entities_mapping.references:
                if entity.origin_type == "imported":
                    module_path = (
                        entity.referenced_entity.module
                        if isinstance(entity.referenced_entity, Import)
                        else os.path.dirname(entity.file_path)
                    )
                    resolved_entities = self.get_entities(entity.name, module_path)

                    if resolved_entities:
                        entity.resolved_entity = resolved_entities[0]
                        self.usages_index.setdefault(
                            f"{entity.resolved_entity.file_path}:{entity.resolved_entity.name}",
                            [],
                        ).append(entity)
                        # logger.debug(f"{entity.full_reference} -> {entity.resolved_entity.full_reference}")
                        entity.origin_type = "imported"  # for go lang
                    else:
                        # logger.warning(f"Could not resolve {entity.full_name} from {entity.file_denotation}")
                        if entity.language == "go" and not entity.object:
                            entity.origin_type = "unknown"  # for go lang
                elif entity.referenced_entity:
                    self.usages_index.setdefault(
                        f"{entity.referenced_entity.file_path}:{entity.referenced_entity.name}",
                        [],
                    ).append(entity)
        # breakpoint()

    def retrieve_entities(self, query: str, include_imports: bool = True):
        # This one loops through get_entities and tries many different variations
        file_path, entity_name = parse_potential_entity_from_line(query)
        *path, entity_name = entity_name.split(".")
        popped_items = ""
        while path:
            current_entity_name = ".".join(path + [entity_name])
            if entities := self.get_entities(
                current_entity_name,
                file_path + popped_items,
                include_imports=include_imports,
            ):
                return entities
            popped_items += "." + path.pop()
        return self.get_entities(entity_name, file_path + popped_items, include_imports=include_imports)

    def get_entities(self, name: str, module_path: str, include_imports: bool = True) -> list[Definition]:
        # This one checks the name and sorts by module_path
        if cached_entities := self.entities_search_cache.get((name, module_path), None):
            return cached_entities
        entities = self.definitions_index.get(name, [])
        if not include_imports:
            entities = [entity for entity in entities if entity.entity_type != "import"]
        # LCS handles the case where the module path is a substring of the file path
        from rapidfuzz.distance import LCSseq  # lazy import for speedup

        entities = sorted(
            entities,
            key=lambda x: (
                x.entity_type != "import",  # de-prioritize imports
                LCSseq.similarity(x.file_path, module_path) if module_path else 0,
                len(self.usages_index.get(f"{x.file_path}:{x.name}", [])),
            ),
            reverse=True,
        )
        # TODO: refactor this out
        self.entities_search_cache[(name, module_path)] = entities
        return entities

    def get_dependencies(self, entity: Entity):
        file_entities = self.entities_mapping[entity.file_path]
        dependencies: list[Reference] = []
        for reference in file_entities.references:
            if reference.start_line >= entity.start_line and reference.end_line <= entity.end_line:
                dependencies.append(reference)
        return dependencies

    def get_entities_data(self, symbol: str, file_path: str = "") -> dict[str, list[Entity]]:
        definitions = self.retrieve_entities(f"{file_path}:{symbol}")
        return {
            "definitions": definitions,
            "usages": self.usages_index.get(f"{file_path}:{symbol}", []),
            "dependencies": (self.get_dependencies(definitions[0]) if definitions else []),
        }

    def hydrate(self, dir_path: str):
        new_entities_mapping = {}
        for file_path, file_entities in tqdm(self.entities_mapping.items(), total=len(self.entities_mapping)):
            try:
                file_contents = open(os.path.join(dir_path, file_path)).read()
            except FileNotFoundError:
                continue
            file_entities.hydrate(file_contents)
            new_entities_mapping[file_path] = file_entities
        self.entities_mapping = new_entities_mapping
        new_definitions_index = {}
        for entities in tqdm(self.definitions_index.values(), total=len(self.definitions_index)):
            for entity in entities:
                try:
                    file_contents = open(os.path.join(dir_path, entity.file_path)).read()
                except FileNotFoundError:
                    continue
                entity.hydrate(file_contents)
                new_definitions_index.setdefault(entity.name, []).append(entity)
        self.definitions_index = new_definitions_index
        new_usages_index = {}
        for usages in tqdm(self.usages_index.values(), total=len(self.usages_index)):  # this one is a bit slower
            for usage in usages:
                try:
                    file_contents = open(os.path.join(dir_path, usage.file_path)).read()
                except FileNotFoundError:
                    continue
                usage.hydrate(file_contents)
                new_usages_index.setdefault(usage.file_path, []).append(usage)
        self.usages_index = new_usages_index

    def to_dict(self):
        data = super().to_dict()
        data["entities_search_cache"] = {}
        return data

    @classmethod
    def from_dict(cls, data):
        with Timer("loading entities mapping"):
            entities_mapping = {
                file_path: FileEntities.from_dict(entities_mapping)
                for file_path, entities_mapping in tqdm(
                    data["entities_mapping"].items(), desc="Loading entities mapping"
                )
            }
        with Timer("loading remaining"):
            definitions_index = {
                name: [Entity.from_dict(**entity) for entity in entities]
                for name, entities in data["definitions_index"].items()
            }
            usages_index = {
                file_path: [Entity.from_dict(**entity) for entity in entities]
                for file_path, entities in data["usages_index"].items()
            }
            results = cls(
                key=data["key"],
                entities_mapping=entities_mapping,
                definitions_index=definitions_index,
                usages_index=usages_index,
            )
        return results



import base64

import chardet
from github.Repository import Repository


# attempts to decode a file with the following encodings
def read_file_with_fallback_encodings(file_path: str, encodings=["utf-8", "windows-1252", "iso-8859-1"]):
    embedded_null_byte = False
    for encoding in encodings:
        try:
            with open(file_path, "r", encoding=encoding) as file:
                return file.read()
        except UnicodeDecodeError:
            continue
        except ValueError:
            embedded_null_byte = True
            continue
    if embedded_null_byte:
        raise Exception(f"Encountered null byte while decoding {file_path}")
    raise UnicodeDecodeError(f"Could not decode {file_path} with any of the specified encodings: {encodings}")


# attempts to incode a string using the following encodings
def encode_file_with_fallback_encodings(file_contents: str, encodings=["utf-8", "windows-1252", "iso-8859-1"]):
    for encoding in encodings:
        try:
            encoded_file = file_contents.encode(encoding)
            return encoded_file
        except UnicodeEncodeError:
            continue
    encodings_string = ", ".join(encodings)
    raise UnicodeEncodeError(f"Could not encode the file with any of the specified encodings: {encodings_string}")


def safe_decode(repo: Repository, path: str, *args, **kwargs):
    """
    By default, this function will decode the file contents from the repo.
    But if the file > 1MB, we will fetch the raw content and then decode it manually ourselves.
    It's a strange bug that occurs when the file is too large and the GitHub API doesn't decode it properly and returns encoding="none".
    Reference: https://docs.github.com/en/rest/repos/contents?apiVersion=2022-11-28#get-repository-content
    """
    try:
        contents = repo.get_contents(path, *args, **kwargs)
        if contents.encoding == "none":
            blob = repo.get_git_blob(contents.sha)
            detected_encoding = chardet.detect(base64.b64decode(blob.content))["encoding"]
            if detected_encoding is None:
                return None
            else:
                try:
                    return base64.b64decode(blob.content).decode(detected_encoding)
                except UnicodeDecodeError as e:
                    raise e
        return contents.decoded_content.decode("utf-8")
    except Exception as e:
        raise e



from typing import Any, Dict, Literal

from pydantic import BaseModel


class Changes(BaseModel):
    body: Dict[str, str] | None = None

    @property
    def previous_body(self):
        return self.body.get("from") if self.body else ""


class Account(BaseModel):
    id: int
    login: str
    type: str


class Installation(BaseModel):
    id: Any | None = None
    account: Account | None = None


class PREdited(BaseModel):
    class Repository(BaseModel):
        full_name: str

    class PullRequest(BaseModel):
        class User(BaseModel):
            login: str

        html_url: str
        title: str
        body: str | None
        number: int

        user: User
        commits: int = 0
        additions: int = 0
        deletions: int = 0
        changed_files: int = 0

    class Sender(BaseModel):
        login: str

    changes: Changes
    pull_request: PullRequest
    sender: Sender
    repository: Repository
    installation: Installation


class InstallationCreatedRequest(BaseModel):
    class Repository(BaseModel):
        full_name: str

    repositories: list[Repository]
    installation: Installation


class ReposAddedRequest(BaseModel):
    class Repository(BaseModel):
        full_name: str

    repositories_added: list[Repository]
    installation: Installation


class CommentCreatedOrEditedRequest(BaseModel):
    class Comment(BaseModel):
        class User(BaseModel):
            login: str
            type: str

        body: str | None
        original_line: int
        path: str
        diff_hunk: str
        user: User
        id: int

    class PullRequest(BaseModel):
        class Head(BaseModel):
            ref: str

        number: int
        body: str | None
        state: str  # "closed" or "open"
        head: Head
        title: str

    class Repository(BaseModel):
        full_name: str
        description: str | None

    class Sender(BaseModel):
        pass

    action: str
    comment: Comment
    pull_request: PullRequest
    repository: Repository
    sender: Sender
    installation: Installation
    changes: Changes | None = None


class IssueRequest(BaseModel):
    class Issue(BaseModel):
        class User(BaseModel):
            login: str
            type: str

        class Assignee(BaseModel):
            login: str

        class Repository(BaseModel):
            # TODO(sweep): Move this out
            full_name: str
            description: str | None

        class Label(BaseModel):
            name: str

        class PullRequest(BaseModel):
            url: str | None

        title: str
        number: int
        html_url: str
        user: User
        body: str | None
        labels: list[Label]
        assignees: list[Assignee] | None = None
        pull_request: PullRequest | None = None

    action: str
    issue: Issue
    repository: Issue.Repository
    assignee: Issue.Assignee | None = None
    installation: Installation | None = None
    sender: Issue.User


class IssueCommentRequest(IssueRequest):
    class Comment(BaseModel):
        class User(BaseModel):
            login: str
            type: Literal["User", "Bot"]

        user: User
        id: int
        body: str

    comment: Comment
    sender: Comment.User
    changes: Changes | None = None


class PRRequest(BaseModel):
    class PullRequest(BaseModel):
        class User(BaseModel):
            login: str

        class MergedBy(BaseModel):
            login: str

        user: User
        title: str
        merged_by: MergedBy | None
        additions: int = 0
        deletions: int = 0

    class Repository(BaseModel):
        full_name: str

    pull_request: PullRequest
    repository: Repository
    number: int

    installation: Installation


class PRLabeledRequest(BaseModel):
    class PullRequest(BaseModel):
        class User(BaseModel):
            login: str

        class Label(BaseModel):
            name: str

        title: str
        labels: list[Label]

        class MergedBy(BaseModel):
            login: str

        user: User
        merged_by: MergedBy | None
        additions: int = 0
        deletions: int = 0

    class Repository(BaseModel):
        full_name: str

    pull_request: PullRequest
    repository: Repository
    number: int

    installation: Installation


class CheckRunCompleted(BaseModel):
    class CheckRun(BaseModel):
        class PullRequest(BaseModel):
            number: int

        class CheckSuite(BaseModel):
            head_branch: str | None

        conclusion: str
        html_url: str
        pull_requests: list[PullRequest]
        completed_at: str
        check_suite: CheckSuite
        head_sha: str

        @property
        def run_id(self):
            # format is like https://github.com/ORG/REPO_NAME/actions/runs/RUN_ID/jobs/JOB_ID
            return self.html_url.split("/")[-3]

    class Repository(BaseModel):
        full_name: str
        description: str | None

    class Sender(BaseModel):
        login: str

    check_run: CheckRun
    installation: Installation
    repository: Repository
    sender: Sender


class GithubRequest(IssueRequest):
    class Sender(BaseModel):
        login: str

    sender: Sender | None = None



from sweepai.config.client import SweepConfig


def cleaned_rg_output(root_directory: str, sweep_config: SweepConfig, output: str):
    # this function now expects that you ran with the cwd as the root dir for the purposes of file paths
    output = output.replace(f"\n{root_directory}/", "\n")
    result_dict = {}
    for block in output.split("\n\n"):
        if not block.strip():
            continue
        file_path, *contents = block.split("\n")
        if not file_path:
            continue
        if sweep_config.is_file_excluded_aggressive(root_directory, file_path):
            continue
        result_dict[file_path] = "\n".join(contents)
    # serialize contents with a string return type
    result = "\n\n".join([f"{file_path}\n{contents}" for file_path, contents in result_dict.items()])
    return result


# try and find code_snippet inside file_contents given various levels of indentation, and right strip the lines of code
# if successful returns the num of spaces required to find the code match and if we need to rstrip the old code or not
def manual_code_check(file_contents: str, code_snippet: str) -> tuple[int, bool]:
    code_lines = [line for line in code_snippet.split("\n")]
    # special case for single line
    if len(code_lines) == 1:
        file_lines = file_contents.split("\n")
        # check directly
        new_code = code_lines[0]
        # only continue if it is unique, this will then later fail the uniqueness check
        if file_contents.count(new_code) > 1:
            return 0, False
        if new_code in file_contents:
            # now check how many leading whitespaces there are
            for line in file_lines:
                if new_code in line:
                    # unless the code is at the start of the line
                    if line.startswith(new_code):
                        return 0, False
                    return len(line) - len(line.lstrip()), False
        else:
            # now try rstrip if initially the code is not there
            new_code = new_code.rstrip()
            if file_contents.count(new_code) > 1:  # uniqueness check
                return 0, False
            if new_code in file_contents:
                # now check how many leading whitespaces there are
                for line in file_lines:
                    if new_code in line:
                        # unless the code is at the start of the line
                        if line.startswith(new_code):
                            return 0, True
                        return len(line) - len(line.lstrip()), True
        return -1, False

    # assume one indent is two spaces and check max 10 indents
    for indent in range(0, 40, 2):
        new_code_lines = [f"{' ' * indent}{line}" if line.strip() else "" for line in code_lines]
        new_code = "\n".join(new_code_lines)
        if new_code in file_contents:
            return indent, False
    # sometimes llm returns code with trailing whitespace, if we have reached here check again but strip all trailing whitespace
    code_lines = [line.rstrip() for line in code_snippet.split("\n")]
    for indent in range(0, 40, 2):
        new_code_lines = [f"{' ' * indent}{line}" if line.strip() else "" for line in code_lines]
        new_code = "\n".join(new_code_lines)
        if new_code in file_contents:
            return indent, True
    return -1, False



import re

from sweepai.config.server import DEV
from sweepai.core.llm.chat import LATEST_CLAUDE_MODEL, Thread

# TODO: add to on_ticket.py

# BAD ISSUES
# - Renaming a class or file or function or variable
# - Deleting a file or dead code or class
# - Attaching files - sweep cannot access urls such as documents or pdf files. For pdfs, say it's not supported rn. Same for screenshots.
# - vague requests. Sweep will reject poorly specified issues.
# - versioning eg mongo 3 to mongo 7 or python 2 to python 3. maybe we can set a limit for these like one file is fine but not codebase wide refactors
# -
# GOOD ISSUES
# - Adding a unit test
# - Well specified task

issue_validator_instructions_prompt = """# Instructions for Evaluating Issues for Sweep

You are an AI assistant tasked with evaluating GitHub issues to determine if they are suitable for Sweep, an automated code modification tool. Your goal is to assess if the issue is clear, actionable, and within Sweep's capabilities.

## Sweep's Capabilities

Sweep can:
- Access and search the entire codebase to find relevant code
- Make targeted code changes to fix bugs or add features 
- Write unit tests well. If the task relates to unit tests, you should usually pass it.
- Refactor small to medium sized components
- Update API usage across a codebase
- Read GitHub Action logs

Sweep works best with issues that:
- Have clear, specific instructions
- Involve changes to <10 files and <2000 lines of code
- Don't require external resources or command line operations

Reject issues that:
- Can be easily done with a basic find-and-replace operation or a rename tool in the IDE.
- Would likely require huge changes, like "migrate my codebase from Python to Rust". This task is too broad for Sweep.

## Examples

Good issues for Sweep:
- "Fix the bug in utils.py where the sort_items() function is not handling empty lists correctly"
- "Add input validation to the create_user() function in users/views.py to check that the email is valid"
- "Refactor the ProductList component in src/components/ProductList.js to use hooks instead of class components"

Issues that may need clarification:
- "Update the entire codebase to Python 3.9" (too broad)
- "Fix all the failing tests" (needs more specifics)
- "Improve performance" (needs concrete steps)

Bad issues for Sweep:
- "Rename the class User to UserDetails in the file users/models.py"
- "Delete the entire utils folder"

## Your Task

For each issue, analyze whether it's suitable for Sweep. If it's unclear or outside Sweep's capabilities, suggest how it could be improved.

Respond in this format:

<thinking>
Provide your analysis here. Consider:
- Is the issue clear and specific?
- Is it within Sweep's capabilities?
- If not, how could it be improved?
</thinking>

<pass>True or False</pass>

If False, respond to the user:
<response_to_user>
1. Explain what you've detected that makes the issue unsuitable for Sweep.
2. Offer suggestions to improve it. Use a friendly but concise tone.
</response_to_user>"""

issue_validator_system_prompt = """You are an AI assistant tasked with determining whether an issue should be passed on to be handled by Sweep, an AI-powered software engineer."""

issue_validator_user_prompt = (
    """<issue>
{issue}
</issue>\n\n"""
    + issue_validator_instructions_prompt
)


def validate_issue(issue: str) -> tuple[bool, str]:
    """
    Validate the issue and provide instructions for overriding if it fails.
    """
    if DEV:
        return True, ""
    thread: Thread = Thread.from_system_message_string(
        prompt_string=issue_validator_system_prompt,
    )

    response = thread.chat(
        issue_validator_user_prompt.format(issue=issue),
        model=LATEST_CLAUDE_MODEL,
        temperature=0.0,
    )

    if "<pass>False</pass>" in response:
        pattern = "<response_to_user>(.*)</response_to_user>"
        validation_message = re.search(pattern, response, re.DOTALL).group(1).strip()
        override_instructions = "\n\n> [!TIP]\n> If you would like to proceed despite this, you can override the validation by adding the text `OVERRIDE_ISSUE_VALIDATION` anywhere in your issue description."
        return False, validation_message + override_instructions
    return True, ""


if __name__ == "__main__":
    print(validate_issue("Rename the method get_jwt in sweepai/core/github_utils.py to get_jwt_2"))



import os
import re

from github.IssueComment import IssueComment
from github.PullRequest import PullRequest
from github.PullRequestComment import PullRequestComment
from github.Repository import Repository
from loguru import logger

from sweepai.backend.api import fix_issue
from sweepai.backend.api_utils import get_pr_snippets
from sweepai.ci.cleanup_ci import clean_ci_logs
from sweepai.core.entities import UserMessage
from sweepai.core.github_utils import ClonedRepo, get_github_client
from sweepai.dataclasses.chat_agent_context import ChatAgentContext
from sweepai.dataclasses.fix_issue_result import FixIssueResult
from sweepai.ci.fetch_ci import create_hyperlink, get_ci_failure_logs, get_ci_failures
from sweepai.handlers.format_issue_results import format_results
from sweepai.o11y.log_utils import log_to_file, log_warnings_to_file
from sweepai.o11y.posthog_trace import posthog_trace
from sweepai.utils.diff import remove_whitespace_changes
from sweepai.utils.str_utils import code_block, format_comment_prompt_from_files, markdown_list, strip_sweep, blockquote, tail, wrap_xml_tag
from sweepai.utils.ticket_rendering_utils import (
    HR,
    IssueErrorHandler,
    center,
    create_message_marker,
    create_or_edit_sweeps_github_comment,
    generate_revert_pr_comment,
    get_suffix_for_sweep_comment,
    sweeping_png,
)
from sweepai.web.validate_license import validate_license

num_of_snippets_to_query = 30
total_number_of_snippet_tokens = 15_000
num_full_files = 2
num_extended_snippets = 2

ERROR_FORMAT = "❌ {title}\n\nPlease report this on our [community forum](https://community.sweep.dev/)."
SWEEPING_GIF = f"{center(sweeping_png)}\n\n"


def get_original_issue(pull_request: PullRequest):
    pattern = r"Fixes\s+#(?P<issue_number>\d+)"
    if match := re.search(pattern, pull_request.body):
        repo = pull_request.base.repo
        issue_number = match.group("issue_number")
        issue = repo.get_issue(int(issue_number))
        return issue
    return None


def get_actor_from_pull_request(pull_request: PullRequest):
    author = pull_request.user
    if author.type != "Bot":
        return author.login

    for assignee in pull_request.assignees:
        if assignee.type != "Bot":
            return assignee.login
    original_issue = get_original_issue(pull_request)

    if original_issue and original_issue.user.type != "Bot":
        return original_issue.user.login

    return author.login

def summarize_ci_logs(logs_list: list, pull_request: PullRequest, username: str) -> str:
    repo_full_name = pull_request.base.repo.full_name
    
    cloned_repo = ClonedRepo.from_repo_full_name(repo_full_name=repo_full_name, branch=pull_request.head.ref)
    pull_request_metadata = generate_pull_request_metadata(pull_request)
    _, _, pulls_message = get_pr_snippets(
        repo_name=repo_full_name,
        annotations={"pulls": pull_request_metadata},
        cloned_repo=cloned_repo,
        context=ChatAgentContext(use_case="ticket"),
    )
    cleaned_logs = clean_ci_logs(
        username=username,
        error_logs=logs_list[0],
        pulls_message=pulls_message,
    )
    return cleaned_logs

def get_ci_data(
    pull_request: PullRequest,
    username: str,
    installation_id: int,
):
    # Fetching CI data
    _pending_runs, _successful_runs, error_runs = get_ci_failures(pull_request)
    logs_list = get_ci_failure_logs(installation_id, error_runs, pull_request)

    # Cleaning CI logs
    if logs_list:
        joined_logs = markdown_list([create_hyperlink(run) for run in error_runs])
        cleaned_logs = summarize_ci_logs(logs_list, pull_request, username)
        ci_comment_body = f" I'm working on this. I also see the following CI errors:\n\n{code_block(tail(cleaned_logs, max_lines=20))}\n\nSources:\n\n{joined_logs}"
    else:
        cleaned_logs = ""
        ci_comment_body = ""
    
    return cleaned_logs, ci_comment_body


def generate_pull_request_metadata(pull_request: PullRequest):
    file_diffs = [
        {
            "sha": pull_request_file.sha,
            "filename": pull_request_file.filename,
            "status": pull_request_file.status,
            "additions": pull_request_file.additions,
            "deletions": pull_request_file.deletions,
            "changes": pull_request_file.changes,
            "blob_url": pull_request_file.blob_url,
            "raw_url": pull_request_file.raw_url,
            "contents_url": pull_request_file.contents_url,
            "patch": remove_whitespace_changes(
                pull_request_file.patch or ""
            ),  # for one java formatting change, we went from 100k chars to 67k chars
            "previous_filename": pull_request_file.previous_filename,
        }
        for pull_request_file in pull_request.get_files()
    ]
    pull_request_metadata = [
        {
            "number": pull_request.number,
            "repo_name": pull_request.base.repo.name,
            "title": pull_request.title,
            "body": pull_request.body,
            "labels": pull_request.labels,
            "status": pull_request.state,
            "file_diffs": file_diffs,
            "branch": pull_request.head.ref,
        }
    ]
    return pull_request_metadata


def generate_revert_comment_and_format_results(
    repo: Repository,
    fix_issue_result: FixIssueResult,
    fix_string: str,
    tracking_id: str,
):
    assert (
        new_pull_request_number := fix_issue_result.pull_request_data.number
    ) is not None, "New pull request number is None"
    new_pull_request = repo.get_pull(new_pull_request_number)
    commit_sha = new_pull_request.head.sha
    revert_pr_comment = generate_revert_pr_comment(
        commit_message=new_pull_request.title,
        commit_sha=commit_sha,
        pull_request_url=new_pull_request.html_url,
    )
    response = format_results(fix_issue_result, fix_string, repo, tracking_id, is_comment=True)
    return response + revert_pr_comment


@posthog_trace
async def on_comment(
    username: str,
    repo_full_name: str,
    installation_id: int,
    pr_number: int,
    comment_id: int,
    tracking_id: str | None = None,
):
    if not os.environ.get("CLI"):
        assert (
            validate_license()
        ), "License key is invalid or expired. Please contact us at team@sweep.dev to upgrade to an enterprise license."

    with log_to_file(tracking_id), log_warnings_to_file(keywords=["[YELLOW]"], tracking_id=tracking_id):
        access_token, g = get_github_client(installation_id)
        repo: Repository = g.get_repo(repo_full_name)
        pull_request: PullRequest = repo.get_pull(pr_number)
        logger.info(f"URL: {pull_request.html_url}")

        # Try to get the comment as a PullRequestComment first
        sweeps_github_comment = None
        with IssueErrorHandler(
            pull_request.as_issue(),
            error_title="Error while processing comment",
            tracking_id=tracking_id,
        ) as monitor:
            original_user_comment = try_get_issue_or_review_comment(comment_id, pull_request)
            logger.info(f"Original user comment: {original_user_comment.html_url}")

            stripped_comment_body = strip_sweep(original_user_comment.body)
            # Create a reply comment depending on the type of comment
            comment_body = f"{blockquote(stripped_comment_body)}\n\nThanks for the comment, @{username}! I'm working on this. This should take about 5-10 min."
            if isinstance(original_user_comment, PullRequestComment):
                sweeps_github_comment = pull_request.create_review_comment_reply(
                    comment_id,
                    comment_body + get_suffix_for_sweep_comment(tracking_id=tracking_id),
                )
            else:
                sweeps_github_comment = create_or_edit_sweeps_github_comment(
                    issue_or_pr=pull_request,
                    sweeps_message=comment_body,
                    sweeps_github_comment=sweeps_github_comment,
                    tracking_id=tracking_id,
                    message_marker=create_message_marker(original_user_comment.html_url),
                )

            monitor.comment = sweeps_github_comment

            # Fetching CI data
            cleaned_logs, ci_comment_body = get_ci_data(pull_request, username, installation_id)

            if ci_comment_body:
                comment_body = f"{blockquote(stripped_comment_body)}\n\nThanks for the comment, @{username}!" + ci_comment_body + "\n\nThis should take about 5-10 min."

                create_or_edit_sweeps_github_comment(
                    issue_or_pr=pull_request,
                    sweeps_message=comment_body,
                    sweeps_github_comment=sweeps_github_comment,
                    tracking_id=tracking_id,
                )

            ### Running Sweep

            pull_request_head_branch = pull_request.head.ref
            pull_request_url = pull_request.html_url
            query = f"Address the following pull request comment"
            # Case 1: Comment is a review comment we can print comment inline with the diff
            if (
                hasattr(original_user_comment, "path")
                and hasattr(original_user_comment, "position")
                and hasattr(original_user_comment, "diff_hunk")
            ):
                query += format_comment_prompt_from_files(
                    comment_path=original_user_comment.path,
                    comment_position=original_user_comment.position,
                    comment_diff_hunk=original_user_comment.diff_hunk,
                    comment_body=stripped_comment_body,
                )
            # Case 2: Comment is an issue comment we can print the comment as a normal comment
            elif hasattr(original_user_comment, "path"):
                query += f" in the file {original_user_comment.path}:\n{stripped_comment_body}"
            # Case 3: Something went wrong, we can print the comment as a normal comment
            else:
                query += f":\n{stripped_comment_body}"

            if cleaned_logs: # TODO: make this routable, don't always include it
                query += f"\n\nAlso, here are the failing CI logs:\n\n" + wrap_xml_tag(cleaned_logs, "failing_ci_logs")

            # This instantiates all of the pull request context for Sweep, simply passing in the URL is not sufficient
            pull_request_metadata = generate_pull_request_metadata(pull_request)
            # Now we run Sweep
            fix_string = f"# Fixes\n > [{stripped_comment_body}]({original_user_comment.html_url})\n\nIn pull request {pull_request_url}\n\n"
            try:
                fix_issue_result = await fix_issue(
                    repo_full_name=repo_full_name,
                    query=query,
                    access_token=access_token,
                    messages=[UserMessage(content=query)],
                    snippets=[],
                    branch=pull_request_head_branch,
                    username=username,
                    pulls=pull_request_metadata,
                )
                comment_body = f"{blockquote(stripped_comment_body)}\n\nHey, @{username}, I finished working on this. Here are the changes I made:"
                response = comment_body + HR + format_results(fix_issue_result, fix_string, repo, tracking_id, is_comment=True)
                create_or_edit_sweeps_github_comment(
                    issue_or_pr=pull_request,
                    sweeps_message=response,
                    sweeps_github_comment=sweeps_github_comment,
                    tracking_id=tracking_id,
                )
                # TODO: update comment body to be simpler
                response = comment_body + HR + generate_revert_comment_and_format_results(repo, fix_issue_result, fix_string, tracking_id)
                create_or_edit_sweeps_github_comment(
                    issue_or_pr=pull_request,
                    sweeps_message=response,
                    sweeps_github_comment=sweeps_github_comment,
                    tracking_id=tracking_id,
                )
                return fix_issue_result
            except Exception as e:
                create_or_edit_sweeps_github_comment(
                    issue_or_pr=pull_request,
                    sweeps_message=f"Error: {str(e)}",
                    sweeps_github_comment=sweeps_github_comment,
                    tracking_id=tracking_id,
                )
                raise e


def try_get_issue_or_review_comment(comment_id: int, pull_request: PullRequest) -> PullRequestComment | IssueComment:
    try:
        comment: PullRequestComment = pull_request.get_review_comment(comment_id)
    except Exception:
        # If it fails, try to get it as an IssueComment
        try:
            comment: IssueComment = pull_request.get_issue_comment(comment_id)
        except Exception as e:
            # Both attempts failed, notify the user
            error_message = f"Failed to retrieve the comment. Error: {str(e)}"
            pull_request.create_issue_comment(error_message)
            raise ValueError(f"Unable to retrieve comment: {error_message}")
    return comment



from dataclasses import dataclass, field
from typing import Iterator, Tuple

from sweepai.core.entities import Snippet


@dataclass
class SeparatedSnippets:
    tools: list[Snippet] = field(default_factory=list)
    junk: list[Snippet] = field(default_factory=list)
    dependencies: list[Snippet] = field(default_factory=list)
    docs: list[Snippet] = field(default_factory=list)
    tests: list[Snippet] = field(default_factory=list)
    source: list[Snippet] = field(default_factory=list)

    def add_snippet(self, snippet: Snippet, type_name: str):
        if type_name == "tools":
            self.tools.append(snippet)
        elif type_name == "junk":
            self.junk.append(snippet)
        elif type_name == "dependencies":
            self.dependencies.append(snippet)
        elif type_name == "docs":
            self.docs.append(snippet)
        elif type_name == "tests":
            self.tests.append(snippet)
        elif type_name == "source":
            self.source.append(snippet)
        else:
            raise ValueError(f"Unknown type_name: {type_name}")

    def override_list(self, attribute_name: str, new_list: list[Snippet]):
        if hasattr(self, attribute_name):
            setattr(self, attribute_name, new_list)
        else:
            raise AttributeError(f"List type '{attribute_name}' does not exist in SeparatedSnippets")

    def __iter__(self) -> Iterator[Tuple[str, list[Snippet]]]:
        yield "source", self.source
        yield "tests", self.tests
        yield "tools", self.tools
        yield "dependencies", self.dependencies
        # yield "docs", self.docs
        # yield "junk", self.junk
        # we won't yield junk or docs
    
    def __len__(self):
        return 4



from dataclasses import dataclass

from sweepai.core.entities import FileChangeRequest
from sweepai.modify.validate.code_validators import CheckResults
from sweepai.search.agent.entity_search import EntitiesIndex


@dataclass
class LLMState:
    request: str
    visited_snippets: set[str]
    visited_questions: set[str]
    visited_ripgrep: set
    ripgrep_queries: set[str]
    viewed_files: set[str]
    entities_index: EntitiesIndex


@dataclass
class ModifyLLMState:
    request: str
    plan: str
    current_task: str
    user_message_index: int
    user_message_index_chat_logger: int
    done_counter: int  # keep track of how many times the submit_task tool has been called
    fcrs: list[FileChangeRequest]
    previous_attempt: str
    changes_per_fcr: list[int]  # how many old/new code pairs there are per fcr
    completed_changes_per_fcr: list[int]  # how many old/new code pairs have been completed per fcr
    attempt_lazy_change: bool  # whether or not we attempt to bypass the llm call and apply old/new code pair directly
    attempt_count: int  # how many times we have attempted to apply the old/new code pair
    visited_set: set[str]
    status_messages: list[str]
    symbol_to_code_map: dict[
        str, str
    ]  # used to provide shortcuts for the llm when it messes up original_code and we provide it the best match
    initial_check_results: dict[str, CheckResults]



# Source: https://gist.github.com/TACIXAT/c5b2db4a80c812c4b4373b65e179a220
"""Formats an sexpression for pretty printing."""


def format_sexpression(s, indent_level=0, indent_size=2):
    """ChatGPT + TACIXAT"""
    output = ""
    i = 0
    # Initialize to False to avoid newline for the first token
    need_newline = False
    cdepth = []  # Track colons
    while i < len(s):
        if s[i] == "(":
            output += "\n" + " " * (indent_level * indent_size) + "("
            indent_level += 1
            need_newline = False  # Avoid newline after opening parenthesis
        elif s[i] == ":":
            indent_level += 1
            cdepth.append(indent_level)  # Store depth where we saw colon
            output += ":"
        elif s[i] == ")":
            indent_level -= 1
            if len(cdepth) > 0 and indent_level == cdepth[-1]:
                # Unindent when we return to the depth we saw the last colon
                cdepth.pop()
                indent_level -= 1
            output += ")"
            need_newline = True  # Newline needed after closing parenthesis
        elif s[i] == " ":
            output += " "
        else:
            j = i
            while j < len(s) and s[j] not in ["(", ")", " ", ":"]:
                j += 1
            # Add newline and indentation only when needed
            if need_newline:
                output += "\n" + " " * (indent_level * indent_size)
            output += s[i:j]
            i = j - 1
            need_newline = True  # Next token should start on a new line
        i += 1
    return output



import base64
import re

import requests
from loguru import logger
from openai import OpenAI

from sweepai.config.client import SweepConfig
from sweepai.core.github_utils import get_installation_id, get_token


# we must get the raw issue which contains the body html and the actual links to the images
def get_image_urls_from_issue(num: int, repo_full_name: str, installation_id: int):
    sweep_config = SweepConfig()
    token = get_token(installation_id)
    url = f"https://api.github.com/repos/{repo_full_name}/issues/{num}"
    headers = {
        "Accept": "application/vnd.github.full+json",
        "Authorization": "Bearer " + token,
        "X-GitHub-Api-Version": "2022-11-28",
    }
    urls = {}
    try:
        response = requests.get(url, headers=headers)
        body_html = response.json()["body_html"]
        if not body_html:
            return urls
        image_url_regex = r'<img.*?src="(?P<url>https?:[^"]+)"'
        image_url_matches = list(re.finditer(image_url_regex, body_html, re.DOTALL))
        for match in image_url_matches:
            url = match.group("url").strip()
            # only accept png, jpg, jpeg or webp
            added = False
            for ext in sweep_config.allowed_image_types:
                if ext in url:
                    urls[url] = ext
                    added = True
                    break
            # unsupported type
            if not added:
                logger.warning(f"Did not add image url: {url}\nReason: image type unsupported!")
    except Exception as e:
        logger.error(f"Encountered error while attempting to fetch raw issue {num} for {repo_full_name}:\n{e}")
    return urls


# gets image contents from a list of urls
def get_image_contents_from_urls(urls: dict[str, str]):
    image_contents = {}
    for url, image_type in urls.items():
        response = requests.get(url)
        if response.status_code == 200:
            image_contents[url] = {
                "content": base64.b64encode(response.content).decode("utf-8"),
                "type": image_type,
            }
        else:
            logger.error(f"Could not get contents for image: {url}")
    return image_contents


# use this function to get placeholder links to the images (what you get from the body of the issue)
# use this in conjunction with get_image_urls_from_issue to see where to place the images
def get_image_urls_from_issue_body(body: str) -> dict[str, str]:
    urls = {}
    image_url_regex = r"(?P<image_text>!\[.*?\]\((?P<url>https?:[^)]+)\))"
    image_url_matches = list(re.finditer(image_url_regex, body, re.DOTALL))

    for match in image_url_matches:
        urls[match.group("url").strip()] = match.group("image_text").strip()
    return urls


# sumarize images
def summarize_images(images: list[tuple[str, str]]):
    summaries = []
    client = OpenAI()
    for image, image_type, url in images:
        response = client.chat.completions.create(
            model="gpt-4-turbo",
            messages=[
                {
                    "role": "user",
                    "content": [
                        {"type": "text", "text": "What’s in this image?"},
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/{image_type};base64,{image}",
                                "detail": "high",
                            },
                        },
                    ],
                }
            ],
            max_tokens=1024,
        )
        summaries.append(response.choices[0].message.content)

    return summaries


def match_image_url_to_image_contents(image_urls: dict[str, str], image_contents: dict[str, dict[str, str]]):
    image_url_to_image_raw_urls = {}
    for image, image_data in image_contents.items():
        image_type = image_data["type"]
        base_url = image.split(f".{image_type}")[0]
        ids = base_url.split(".com/")[-1]  # this is slightly jank
        id1, id2 = ids.split("/")
        id2 = "-".join(id2.split("-")[1:])
        # match image_url based on this unique_id
        unique_id = f"{id1}/{id2}"
        for image_url in image_urls:
            if unique_id in image_url:
                image_url_to_image_raw_urls[image_url] = image
    return image_url_to_image_raw_urls


# turns a block of text into a message with images
def create_message_with_images(message: dict[str, str], images: dict[str, dict[str, str]], use_openai: bool = False):
    if not images:
        return message
    image_urls = get_image_urls_from_issue_body(message["content"])
    if not image_urls:
        return message
    # we have detected the presence of images, now we need to match our image with the actual image
    image_url_to_image_raw_urls = match_image_url_to_image_contents(image_urls, images)
    # now break the message up
    message_contents = message["content"]
    new_contents = []
    # now we find each place in the message contents our image texts show up
    markers = list(image_urls.items())
    positions = []
    for key, marker in markers:
        for match in re.finditer(re.escape(marker), message_contents):
            positions.append((match.start(), marker, key))
    positions.sort()

    last_pos = 0
    for pos, marker, image_url in positions:
        if image_url not in image_url_to_image_raw_urls:
            continue
        image_data = images[image_url_to_image_raw_urls[image_url]]
        new_contents.append(
            {
                "type": "text",
                "text": message_contents[last_pos:pos] + "\nAn image has been attached below:",
            }
        )
        if use_openai:
            # add our image block
            new_contents.append(
                {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:image/{image_data['type']};base64,{image_data['content']}",
                        "detail": "high",
                    },
                }
            )
        else:
            # add our image block
            new_contents.append(
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": f"image/{image_data['type']}",
                        "data": image_data["content"],
                    },
                }
            )
        last_pos = pos + len(marker)
    # add last bit
    if message_contents[last_pos:].strip():
        new_contents.append({"type": "text", "text": message_contents[last_pos:]})

    new_message = {"role": message["role"], "content": new_contents}
    return new_message


if __name__ == "__main__":
    repo_full_name = "org/repo"
    issue_number = 0
    org_name, repo_name = repo_full_name.split("/")
    installation_id = get_installation_id(org_name)
    image_urls = get_image_urls_from_issue(issue_number, repo_full_name, installation_id)
    image_contents = get_image_contents_from_urls(image_urls)



from __future__ import annotations

import os
import re
from difflib import unified_diff
from typing import Any, ClassVar, Literal, Type, TypeVar

from loguru import logger
from pydantic import BaseModel, Field

from sweepai.utils.diff import generate_diff
from sweepai.utils.str_utils import (
    add_line_numbers,
    blockquote,
    rstrip_lines,
    strip_triple_quotes,
)

Self = TypeVar("Self", bound="RegexMatchableBaseModel")


# pylint: disable=no-member,unsubscriptable-object
class Message(BaseModel):
    role: Literal["system"] | Literal["user"] | Literal["assistant"] | Literal["function"]
    content: str | None = None
    name: str | None = None
    function_call: dict | None = None
    key: str | None = None
    annotations: dict | None = None

    @classmethod
    def from_tuple(cls, tup: tuple[str | None, str | None]) -> Self:
        if tup[0] is None:
            return cls(role="assistant", content=tup[1])
        else:
            return cls(role="user", content=tup[0])

    def to_openai(self) -> str:
        obj = {
            "role": self.role,
            "content": self.content,
        }
        if self.function_call:
            obj["function_call"] = self.function_call
        if self.role == "function":
            obj["name"] = self.name
        return obj

    def __repr__(self):
        # take the first 100 and last 100 characters of the message if it's too long
        if self.content is None:
            return "No content."
        truncated_message_content = (
            self.content[:100] + "..." + self.content[-100:] if len(self.content) > 200 else self.content
        )
        return f"\nSTART OF MESSAGE\n\n{truncated_message_content}\n\nROLE: {self.role} FUNCTION_CALL: {self.function_call} NAME: {self.name} ANNOTATIONS: {self.annotations if self.annotations else ''} KEY: {self.key}\n\nEND OF MESSAGE\n\n"


class SystemMessage(Message):
    role: Literal["system"] = "system"
    key: str = "system"


class UserMessage(Message):
    role: Literal["user"] = "user"


class AssistantMessage(Message):
    role: Literal["assistant"] = "assistant"


class FunctionMessage(Message):
    role: Literal["function"] = "function"


class Function(BaseModel):
    class Parameters(BaseModel):
        type: str = "object"
        properties: dict

    name: str
    description: str
    parameters: Parameters


class RegexMatchError(ValueError):
    pass


class RegexMatchableBaseModel(BaseModel):
    _regex: ClassVar[str]

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        # match = re.search(file_regex, string, re.DOTALL)
        match = re.search(cls._regex, string, re.DOTALL)
        if match is None:
            logger.warning(f"Did not match {string} with pattern {cls._regex}")
            raise RegexMatchError("Did not match")
        return cls(
            **{k: (v if v else "").strip("\n") for k, v in match.groupdict().items()},
            **kwargs,
        )


class FileChangeRequest(RegexMatchableBaseModel):
    filename: str
    instructions: str
    change_type: (
        Literal["modify"]
        | Literal["create"]
        | Literal["delete"]
        | Literal["rename"]
        | Literal["rewrite"]
        | Literal["check"]
        | Literal["refactor"]
        | Literal["test"]
    )
    _regex = r"""<(?P<change_type>[a-z_]+)\s+file=\"(?P<filename>[a-zA-Z0-9/\\\.\[\]\(\)\_\+\- @\{\}]*?)\"( start_line=\"(?P<start_line>.*?)\")?( end_line=\"(?P<end_line>.*?)\")?( entity=\"(.*?)\")?( source_file=\"(?P<source_file>.*?)\")?( destination_module=\"(?P<destination_module>.*?)\")?( relevant_files=\"(?P<raw_relevant_files>.*?)\")?(.*?)>(?P<instructions>.*?)\s*<\/\1>"""
    is_completed: bool = False
    entity: str | None = None
    source_file: str | None = None
    old_content: str | None = None
    new_content: str | None = None
    raw_relevant_files: str | None = None
    # allow inf
    start_line: Any | int | str | None = None
    end_line: Any | int | str | None = None
    start_and_end_lines: list[tuple] = []
    comment_line: int | None = None
    sandbox_response: None = None
    failed_sandbox_test: bool | None = False
    parent: FileChangeRequest | None = None
    status: Literal["succeeded"] | Literal["failed"] | Literal["queued"] | Literal["running"] = "queued"
    destination_module: str | None = None
    commit_hash_url: str | None = None
    id: str | None = None

    def __repr__(self):
        return f"START OF FCR\n\n{self.change_type.capitalize()}: {self.filename} with instructions:\n{self.instructions}\n\nEND OF FCR\n\n"

    def get_edit_url(self, repo_full_name: str, branch_name: str):
        url = f"https://github.com/{repo_full_name}/edit/{branch_name}/{self.filename}"
        if self.start_line and self.end_line:
            url += f"#L{self.start_line}-L{self.end_line}"
        return url

    @classmethod
    def from_string(cls: Type[Self], string: str, **kwargs) -> Self:
        result = super().from_string(string, **kwargs)
        result.filename = result.filename.strip("/")
        result.instructions = result.instructions.replace("\n*", "\n•")
        if result.source_file:
            result.source_file = result.source_file.strip()
            if " " in result.source_file:
                result.source_file = result.source_file.split(" ")[0]
        if result.instructions.startswith("*"):
            result.instructions = "•" + result.instructions[1:]
        if result.start_line:
            try:
                result.start_line = int(result.start_line)
            except ValueError:
                result.start_line = None
        if result.end_line:
            try:
                result.end_line = int(result.end_line)
            except ValueError:
                result.start_line = None
        return result

    @property
    def relevant_files(self):
        if not self.raw_relevant_files:
            return []

        return [relevant_file for relevant_file in self.raw_relevant_files.split(" ") if relevant_file != self.filename]

    @property
    def entity_display(self):
        if self.entity:
            return f"`{self.filename}:{self.entity}`"
        else:
            return f"`{self.filename}`"

    @property
    def status_display(self):
        if self.status == "succeeded":
            return "✓"
        elif self.status == "failed":
            if self.change_type == "modify":
                return "! No changes made"
            return "✗"
        elif self.status == "queued":
            return "▶"
        elif self.status == "running":
            return "⋯"
        else:
            raise ValueError(f"Unknown status {self.status}")

    @property
    def display_summary(self):
        if self.change_type == "check":
            return f"Running GitHub Actions for `{self.filename}`"
        return f"{self.change_type.capitalize()} `{self.filename}`"

    @property
    def summary(self):
        prefix = {"failed": "✗", "succeeded": "✓", "queued": "▶", "running": "⋯"}[self.status] + " "
        return prefix + f"{self.change_type.capitalize()}\n{self.filename}"

    @property
    def color(self):
        color_map = {
            "failed": "red2",
            "succeeded": "#0ee832",
            "queued": "white",
            "running": "yellow",
        }
        return color_map[self.status]

    @property
    def entity_display_without_backtick(self):
        if self.entity:
            return f"{self.filename}:{self.entity}"
        else:
            return f"{self.filename}"

    @property
    def instructions_ticket_display(self):
        return self.instructions_display

    @property
    def instructions_display(self):
        # if self.change_type == "check":
        #     return f"Run GitHub Actions for `{self.filename}` with results:\n{self.instructions}"
        return f"{self.change_type.capitalize()} {self.filename} with contents:\n{self.instructions}"

    @property
    def diff_display(self):
        if self.old_content and self.new_content:
            diff = unified_diff(
                self.old_content.splitlines(keepends=True),
                self.new_content.splitlines(keepends=True),
            )
            diff_text = "".join(diff)
            return f"<pre>{diff_text}</pre>"
        return ""


class SweepPullRequest(RegexMatchableBaseModel):
    title: str
    branch_name: str
    content: str
    _regex = r'''pr_title\s+=\s+"(?P<title>.*?)"\n+branch\s+=\s+"(?P<branch_name>.*?)"\n+pr_content\s+=\s+f?"""(?P<content>.*?)"""'''


class ProposedIssue(RegexMatchableBaseModel):
    title: str
    body: str
    issue_id: int | None = None
    _regex = r'<issue\s+title="(?P<title>.*?)">(?P<body>.*?)</issue>'


SNIPPET_FORMAT = """<code_block>
<source>{denotation}</source>
<code_block_content>
{contents}
</code_block_content>
</code_block>"""


# pylint: disable=no-member,unsubscriptable-object
class Snippet(BaseModel):
    """
    Start and end refer to line numbers
    """

    content: str = Field(repr=False)
    start: int
    end: int
    file_path: str
    score: float = 0.0  # TODO: migrate all usages to use this
    type_name: Literal["source", "tests", "dependencies", "tools", "docs"] = "source"

    def __init__(self, **data):
        super().__init__(**data)
        self.start = max(1, self.start)

    def __eq__(self, other):
        if isinstance(other, Snippet):
            return self.file_path == other.file_path and self.start == other.start and self.end == other.end
        return False

    def __hash__(self):
        return hash((self.file_path, self.start, self.end))

    def __repr__(self):
        return f"{self.file_path}:{self.start}-{self.end}\n"

    def get_snippet(self, add_ellipsis: bool = True, add_lines: bool = True):
        if self.content is None:
            return ""
        lines = self.content.splitlines()
        snippet = "\n".join(lines[max(self.start - 1, 0) : self.end])

        if add_lines:
            snippet = add_line_numbers(snippet, start=self.start)

        if add_ellipsis:
            if self.start > 1:
                snippet = "...\n" + snippet
            if self.end < self.content.count("\n") + 1:
                snippet = snippet + "\n..."
        return snippet

    def __add__(self, other):
        assert self.content == other.content
        assert self.file_path == other.file_path
        return Snippet(
            content=self.content,
            start=self.start,
            end=other.end,
            file_path=self.file_path,
        )

    def __xor__(self, other: "Snippet") -> bool:
        """
        Returns True if there is an overlap between two snippets.
        """
        if self.file_path != other.file_path:
            return False
        return self.file_path == other.file_path and (
            (self.start <= other.start and self.end >= other.start)
            or (other.start <= self.start and other.end >= self.start)
        )

    def __or__(self, other: "Snippet") -> "Snippet":
        assert self.file_path == other.file_path
        return Snippet(
            content=self.content,
            start=min(self.start, other.start),
            end=max(self.end, other.end),
            file_path=self.file_path,
        )

    @property
    def xml(self):
        return SNIPPET_FORMAT.format(denotation=self.file_denotation, contents=self.get_snippet())

    def expand(self, num_lines: int = 25):
        return Snippet(
            content=self.content,
            start=max(self.start - num_lines, 1),
            end=min(self.end + num_lines, self.content.count("\n") + 1),
            file_path=self.file_path,
            score=self.score,
        )

    def render(self, i: int | None = None, EXPAND_SIZE: int = 100):
        headers = '<relevant_file index="' + str(i) + '">' if i is not None else "<relevant_file>"
        return f"""{headers}
<file_path>
{self.file_denotation}
</file_path>
<source>
{self.expand(EXPAND_SIZE).get_snippet(add_lines=False)}
</source>
</relevant_file>"""

    @property
    def denotation(self):
        return f"{self.file_path}:{self.start}-{self.end}"

    @property
    def file_denotation(self):
        if self.start <= 0 and self.end >= self.content.count("\n") + 1:
            return f"{self.file_path}"
        return f"{self.file_path}:{self.start}-{self.end}"

    @classmethod
    def from_file(cls, file_path: str, file_contents: str, **kwargs):
        return cls(
            content=file_contents,
            start=0,
            end=file_contents.count("\n"),
            file_path=file_path,
            **kwargs,
        )

def sort_and_fuse_snippets(
    snippets: list[Snippet],
    fuse_distance: int = 600,
) -> list[Snippet]:
    if len(snippets) <= 1:
        return snippets
    new_snippets = []
    snippets.sort(key=lambda x: x.start)
    current_snippet = snippets[0]
    for snippet in snippets[1:]:
        if current_snippet.end + fuse_distance >= snippet.start:
            current_snippet.end = max(current_snippet.end, snippet.end)
            current_snippet.score = max(current_snippet.score, snippet.score)
        else:
            new_snippets.append(current_snippet)
            current_snippet = snippet
    new_snippets.append(current_snippet)
    return new_snippets


def organize_snippets(snippets: list[Snippet], fuse_distance: int = 600) -> list[Snippet]:
    """
    Fuse and dedup snippets that are contiguous. Combine ones of same file.
    """
    fused_snippets = []
    added_file_paths = set()
    for i, snippet in enumerate(snippets):
        if snippet.file_path in added_file_paths:
            continue
        added_file_paths.add(snippet.file_path)
        current_snippets = [snippet]
        for current_snippet in snippets[i + 1 :]:
            if snippet.file_path == current_snippet.file_path:
                current_snippets.append(current_snippet)
        current_snippets = sort_and_fuse_snippets(current_snippets, fuse_distance=fuse_distance)
        fused_snippets.extend(current_snippets)
    return fused_snippets

def snippet_to_string(snippet: Snippet) -> str:
    return snippet.get_snippet(add_lines=False, add_ellipsis=False)


def fuse_snippets(snippets: list[Snippet]) -> list[Snippet]:
    """
    Fuses overlapping snippets into a single snippet.
    """
    new_snippets: list[Snippet] = []
    for snippet in snippets:
        for new_snippet in new_snippets:
            if new_snippet.file_path == snippet.file_path:
                if new_snippet.end + 1 == snippet.start:
                    new_snippet.end = snippet.end
                    break
                elif new_snippet.start - 1 == snippet.end:
                    new_snippet.start = snippet.start
                    break
        else:
            new_snippets.append(snippet)
    return new_snippets


class NoFilesException(Exception):
    def __init__(self, message="Sweep could not find any files to modify"):
        super().__init__(message)


class UnsuitableFileException(Exception):
    def __init__(self, message="Sweep has determined this file is unsuitable to work with"):
        super().__init__(message)


class PRChangeRequest(BaseModel):
    params: dict


class MockPR(BaseModel):
    # Used to mock a PR object without creating a PR (branch will be created tho)
    file_count: int = 0  # Number of files changes
    title: str
    body: str
    pr_head: str
    base: Any
    head: Any
    assignee: Any = None

    id: int = -1
    state: str = "open"
    html_url: str = ""

    def create_review(self, *args, **kwargs):
        # Todo: used to prevent erroring in on_review.py file
        pass

    def create_issue_comment(self, *args, **kwargs):
        pass


class SandboxResponse(BaseModel):
    success: bool
    outputs: list[str]
    updated_content: str
    error_messages: list[str]


class MaxTokensExceeded(Exception):
    def __init__(self, filename):
        self.filename = filename


class UnneededEditError(Exception):
    def __init__(self, filename):
        self.filename = filename


class MatchingError(Exception):
    def __init__(self, filename):
        self.filename = filename


class EmptyRepository(Exception):
    def __init__(self):
        pass


def parse_fcr(fcr: FileChangeRequest):
    justification, *_ = fcr.instructions.split("<original_code>", 1)
    justification, *_ = justification.split("<new_code>", 1)
    justification = (
        justification.rstrip().removesuffix("1.").removesuffix("2.").rstrip()
    )  # sometimes Claude puts 1. <original_code> which is weird
    original_code_pattern = r"<original_code(?: file_path=\".*?\")?(?: index=\"\d+\")?>\s*\n(.*?)</original_code>"
    new_code_pattern = r"<new_code(?: file_path=\".*?\")?(?: index=\"\d+\")?>\s*\n(.*?)</new_code>"
    original_code_matches = list(re.finditer(original_code_pattern, fcr.instructions, re.DOTALL))
    new_code_matches = list(re.finditer(new_code_pattern, fcr.instructions, re.DOTALL))
    replace_all_pattern = r"<replace_all>true</replace_all>"
    replace_all_matches = list(re.finditer(replace_all_pattern, fcr.instructions, re.DOTALL))
    return {
        "justification": justification.strip(),
        "file_path": fcr.filename,
        "original_code": [
            strip_triple_quotes(original_code_match.group(1)) for original_code_match in original_code_matches
        ],
        "new_code": [strip_triple_quotes(new_code_match.group(1)) for new_code_match in new_code_matches],
        "replace_all": bool(replace_all_matches),
    }


def render_fcrs(file_change_requests: list[FileChangeRequest]):
    # Render plan start
    planning_markdown = ""
    for fcr in file_change_requests:
        parsed_fcr = parse_fcr(fcr)
        if parsed_fcr and parsed_fcr["new_code"]:
            planning_markdown += f"#### `{fcr.filename}`\n"
            planning_markdown += f"{blockquote(parsed_fcr['justification'])}\n\n"
            if parsed_fcr["original_code"] and parsed_fcr["original_code"][0].strip():
                planning_markdown += f"""```diff\n{generate_diff(
                    parsed_fcr["original_code"][0],
                    rstrip_lines(parsed_fcr["new_code"][0]),
                )}\n```\n"""
            else:
                _file_base_name, ext = os.path.splitext(fcr.filename)
                planning_markdown += f"```{ext}\n{parsed_fcr['new_code'][0]}\n```\n"
        else:
            planning_markdown += f"#### `{fcr.filename}`\n{blockquote(fcr.instructions)}\n"
    return planning_markdown


if __name__ == "__main__":
    snippet = Snippet(content="test", start=-10, end=10, file_path="test.py", score=1.0)
    print(snippet)
    # bp()



import os

from anytree import Node, RenderTree
from diskcache import Cache
from loguru import logger
from tqdm import tqdm

from sweepai.config.client import SweepConfig
from sweepai.config.server import CACHE_DIRECTORY
from sweepai.core.entities import SNIPPET_FORMAT, Snippet
from sweepai.core.github_utils import ClonedRepo
from sweepai.core.llm.chat import LATEST_CLAUDE_MODEL, call_llm
from sweepai.search.query.repo_parsing_utils import is_actual_code_file
from sweepai.utils.str_utils import extract_xml_tag, pack_items_for_prompt
from sweepai.utils.timer import Timer

summary_caches = Cache(f"{CACHE_DIRECTORY}/summary_caches")

FILE_THRESHOLD = 500


def estimate_line_count(filename):
    stat = os.stat(filename)
    # Get file size
    file_size = stat.st_size
    # Read at most 8000 bytes, ~160 lines of python code to estimate average line length.
    # Obvious limitations here but it's a good enough heuristic for now. Overestimates for short lines.
    sample_size = min(file_size, 8000)
    with open(filename, "rb") as f:
        sample = f.read(sample_size)
    # Count newlines in the sample
    sample_newlines = sample.count(b"\n")
    # Estimate average line length
    if sample_newlines == 0:
        return 1  # Assume it's a one-line file if no newlines in sample
    avg_line_length = len(sample) / sample_newlines
    # Estimate total lines
    estimated_lines = int(file_size / avg_line_length)
    return estimated_lines


def count_descendants_files_or_lines_of_code(directory: str, count_lines_of_code: bool = False):
    sweep_config = SweepConfig()
    descendant_count = {}
    dir_file_count = {}

    def is_dir_too_big(file_name: str):
        dir_name = os.path.dirname(file_name)
        file_sections = file_name.split(os.sep)
        # use filter_file here
        for name in (
            ".git",
            "node_modules",
            ".venv",
            "build",
            "venv",
            "patch",
            ".next",
        ):
            if name in file_sections:
                return True
        if dir_name not in dir_file_count:
            dir_file_count[dir_name] = len(os.listdir(dir_name))
        return dir_file_count[dir_name] > FILE_THRESHOLD

    def dfs(current_dir):
        count = 0
        for root, sub_paths, files in os.walk(current_dir):
            if is_dir_too_big(root):
                continue
            for sub_path in sub_paths:
                if is_dir_too_big(os.path.join(root, sub_path)):
                    continue
                count += dfs(os.path.join(root, sub_path))
            if count_lines_of_code:
                for file in files:
                    if is_actual_code_file(root, os.path.join(root, file), sweep_config):
                        # add number of estimated newlines in file
                        count += estimate_line_count(os.path.join(root, file))
            else:
                count += len(files)
        descendant_count[current_dir.removeprefix(directory.rstrip() + "/")] = count
        return count

    print("Counting descendants")
    with Timer():
        dfs(directory)
    print("Done counting descendants")
    for key in (".git", ""):
        if key in descendant_count:
            del descendant_count[key]
    descendant_count["/"] = descendant_count[directory]
    del descendant_count[directory]
    return descendant_count


# include files, and important structural information from the readme
instructions = "Start by explaining the high-level purpose of the directory and its subdirectories. Focus on nuances and patterns. Then, explain what types of business logic this directory contains, with examples from specific files and directories. Only list describe contents that appear in multiple files. Be concise and optimize for informational density. One paragraph."

system_prompt = "Your job is to summarize the following directory from the repository. " + instructions

user_prompt = (
    """Summarize the following directory from the repository.

Repository:
{repo_name}

Directory:
{directory}

<example_files>
{snippets_string}
</example_files>

"""
    + instructions
)


def summarize_directory(
    directory: str,
    snippets: list[Snippet],
    cloned_repo: ClonedRepo,
    directory_summaries: dict[str, str] = {},
):
    key = (
        "summary",
        cloned_repo.repo_full_name,
        "\n".join([snippet.denotation for snippet in snippets]),
    )
    if key in summary_caches:
        return summary_caches[key]
    directory_files = sorted(
        [
            subdir
            for subdir in os.listdir(os.path.join(cloned_repo.repo_dir, directory))
            if os.path.isfile(os.path.join(cloned_repo.repo_dir, directory, subdir))
        ]
    )
    directory_subdirs = sorted(
        [
            subdir
            for subdir in os.listdir(os.path.join(cloned_repo.repo_dir, directory))
            if os.path.isdir(os.path.join(cloned_repo.repo_dir, directory, subdir))
        ]
    )

    snippets_string = "\n\n".join(
        [
            SNIPPET_FORMAT.format(
                denotation=snippet.denotation,
                contents=snippet.expand(50).get_snippet(False, False),
                index=index,
            )
            for index, snippet in enumerate(snippets)
        ]
    )

    other_files_string = ""
    for file in directory_files:
        if file not in [snippet.file_path for snippet in snippets]:
            other_files_string += "- " + file + "\n"
    if other_files_string:
        snippets_string += f"\n\nThe directory also contains the following files:\n\n" + other_files_string

    for subdir, summary in directory_summaries.items():
        if subdir.startswith(directory) and summary:
            snippets_string += f"\n\nHere is a summary of the subdirectory {subdir}:\n\n" + summary

    subdirs_string = ""
    for subdir in directory_subdirs:
        if subdir not in directory_summaries:
            subdirs_string += "- " + subdir + "\n"
    if subdirs_string:
        snippets_string += f"\n\nThe directory also contains the following subdirectories:\n\n" + subdirs_string

    response = call_llm(
        system_prompt,
        user_prompt,
        params={
            "repo_name": cloned_repo.repo_full_name,
            "directory": directory,
            "snippets_string": snippets_string.strip(),
        },
        verbose=False,
    )
    summary_caches[key] = response

    return response


NUM_SNIPPET_EXAMPLES = 10


def get_last_commit_time(cloned_repo: ClonedRepo, num_commits: int = 1000):
    # loop through last 1000 commits
    key = ("last_commit_time", cloned_repo.repo_full_name)
    if key in summary_caches:
        return summary_caches[key]
    last_commit_time = {}
    for commit in tqdm(
        cloned_repo.git_repo.iter_commits(max_count=num_commits),
        desc=f"Getting last commit time for last {num_commits} commits",
        total=num_commits,
    ):
        for file_path in commit.stats.files:
            last_commit_time[file_path] = commit.committed_date
    summary_caches[key] = last_commit_time
    return last_commit_time


def recursively_summarize_directory(cloned_repo: ClonedRepo, update_summaries: bool = False):
    key = ("summaries", cloned_repo.repo_full_name)
    if not update_summaries and key in summary_caches:  # let's just read from the cache
        summaries, counts = summary_caches[key]
        return summaries, counts

    last_commit_time = get_last_commit_time(cloned_repo)

    sweep_config = SweepConfig()
    directory = cloned_repo.repo_dir
    descendants_lines_of_code = count_descendants_files_or_lines_of_code(directory, count_lines_of_code=True)
    directory_summaries = {}
    # go in reverse order
    for subdir in tqdm(sorted(descendants_lines_of_code.keys(), key=lambda x: x.count("/"), reverse=True)):
        # check lines of code in directory instead of descendant counts, sometimes there will be a lot of useful logic in a directory but not many files
        if descendants_lines_of_code[subdir] <= 300:
            continue
        logger.info(f"Summarizing {subdir}")
        snippets_in_subdir = []
        for file_path in os.listdir(os.path.join(cloned_repo.repo_dir, subdir)):
            if os.path.isdir(os.path.join(cloned_repo.repo_dir, subdir, file_path)):
                continue
            try:
                if not is_actual_code_file(
                    os.path.join(cloned_repo.repo_dir, subdir),
                    os.path.join(cloned_repo.repo_dir, subdir, file_path),
                    sweep_config,
                ):
                    continue
                file_contents = open(os.path.join(cloned_repo.repo_dir, subdir, file_path)).read()
                if file_contents.count("\n") > 5000:  # decent filter for now
                    continue
                snippets_in_subdir.append(
                    Snippet(
                        content=file_contents,
                        start=0,
                        end=file_contents.count("\n") + 1,
                        file_path=file_path,
                    )
                )
            except Exception:
                continue

        def snippet_cmp(x):
            return (
                x.file_path.endswith("README.md"),
                x.file_path.endswith(".md"),
                x.file_path.endswith(".rst"),
                last_commit_time.get(file_path, 0),
                x.file_path,
            )

        # put readme first, then md files, then rst, then commit date, then others
        snippets_in_subdir = sorted(snippets_in_subdir, key=snippet_cmp, reverse=True)
        # approximate context for snippets rather than naive slicing
        snippets_in_subdir = pack_items_for_prompt(
            iterable=snippets_in_subdir,
            string_function=lambda snippet: snippet.get_snippet(False, False),
            token_limit=50_000,
        )
        # if there's nothing in the directory, don't summarize
        # there was a bug in summarize_directory where it would summarize the directory even if there were no snippets and the cache was wrong
        if not snippets_in_subdir:
            continue
        # removed mute()
        directory_summaries[subdir] = summarize_directory(subdir, snippets_in_subdir, cloned_repo, directory_summaries)
    summary_caches[key] = directory_summaries, descendants_lines_of_code
    return directory_summaries, descendants_lines_of_code


PRIORITIZED_FILES = [
    "README.md",
    "README.rst",
    "README.txt",
    "README",
    "package.json",
    "requirements.txt",
    "setup.py",
    "setup.cfg",
    "pyproject.toml",
    "cargo.toml",
    "go.mod",
]


def fetch_all_prioritized_files(cloned_repo: ClonedRepo):
    prioritized_file_contents = {}
    for root, dirs, files in os.walk(cloned_repo.repo_dir):
        for file in files:
            if file in PRIORITIZED_FILES:
                file_path = os.path.join(root, file)
                relative_path = os.path.relpath(file_path, cloned_repo.repo_dir)
                try:
                    with open(file_path, "r", encoding="utf-8") as f:
                        content = f.read()
                    if len(content.splitlines()) < 5:
                        continue
                    prioritized_file_contents[relative_path] = content
                except Exception as e:
                    logger.error(f"Error reading file {file_path}: {str(e)}")

    return prioritized_file_contents


# expanded directories
def generate_tree(cloned_repo: ClonedRepo, max_depth: int = 3, max_files: int = 15):
    descendant_counts = count_descendants_files_or_lines_of_code(cloned_repo.repo_dir)
    root = Node(os.path.basename(cloned_repo.repo_dir))

    def add_children(parent, path, depth=0):
        if depth >= max_depth:
            relative_path = path.removeprefix(cloned_repo.repo_dir).lstrip("/")
            try:
                parent.name += f"/ ({descendant_counts[relative_path]} files)"
            except Exception:
                pass
            return
        directories = []
        files = []
        for item in os.listdir(path):
            item_path = os.path.join(path, item)
            if os.path.isdir(item_path):
                if item not in [".git", "node_modules", ".venv", "build", "venv"]:
                    directories.append(item)
            else:
                files.append(item)

        for directory in sorted(directories):
            child = Node(directory, parent=parent)
            add_children(child, os.path.join(path, directory), depth + 1)

        sorted_files = sorted(
            files,
            key=lambda x: (
                PRIORITIZED_FILES.index(x) if x in PRIORITIZED_FILES else 1000,
                x,
            ),
        )
        if depth == 0 or any(x.split("/")[-1] in PRIORITIZED_FILES for x in sorted_files):  # For root directories
            for file in sorted_files:
                Node(file, parent=parent)
        else:
            for i, file in enumerate(sorted_files):
                if i < max_files:
                    Node(file, parent=parent)
                elif i == max_files:
                    remaining = len(sorted_files) - max_files
                    Node(f"... {remaining} more files", parent=parent)
                    break

    add_children(root, cloned_repo.repo_dir)

    tree_string = ""
    for pre, _, node in RenderTree(root):
        tree_string += f"{pre}{node.name}\n"

    return tree_string


frameworks_format_prompt = """Follow this XML-based format:

<answer>
<summary>
Summary of the repo's purpose and its contents. Then list all significant directories and subdirectories and their purpose.
</summary>

<stack>
Frontend (path/to/directory):
- Languages: Languages used in descending order of frequency
- Frameworks:
    - Framework for purpose
    [list all frameworks used, such as for components, styling, etc.]
- Testing:
    - Framework for testing, mocks, etc.

Backend (path/to/directory):
- Languages: Languages used in descending order of frequency
- Frameworks:
    - Framework for purpose
    [list all frameworks used, such as for routing, database, authentication, caching, observability etc.]
- Testing:
    - Framework for testing, mocks, etc.

[repeat for mobile, infrastructure, jobs, microservices, and any other miscellaneous tools and relevant third-party services etc.]
</stack>
</answer>"""

frameworks_system_prompt = (
    """You are an expert at summarizing the tooling and frameworks used in a project.

For each project of the directory, such as the frontend, mobile, backend or microservices etc., list out all languages and frameworks used in the source code, their corresponding purpose and any frameworks used for their testing.

"""
    + frameworks_format_prompt
)


def summarize_frameworks(
    cloned_repo: ClonedRepo,
):
    key = ("frameworks", cloned_repo.repo_full_name, "v0.0")
    if key in summary_caches:
        return summary_caches[key]
    context = ""
    tree = generate_tree(cloned_repo)
    context += tree
    prioritized_file_contents = fetch_all_prioritized_files(cloned_repo)
    for file_path, content in prioritized_file_contents.items():
        context += f"\n\nFile: {file_path}\n\n<file_contents>\n{content}\n</file_contents>\n\n"

    response = call_llm(
        frameworks_system_prompt,
        context + "\n\n" + frameworks_format_prompt,
        model=LATEST_CLAUDE_MODEL,
    )
    response = extract_xml_tag(response, "answer")
    summary_caches[key] = response, tree
    return response, tree


if __name__ == "__main__":
    from sweepai.core.github_utils import MockClonedRepo
    from sweepai.search.query.lexical_search import prepare_lexical_search_index

    REPO = os.environ.get("REPO")
    org_name, repo_name = REPO.split("/")
    cloned_repo = MockClonedRepo(f"/tmp/{repo_name}", REPO)
    directory = "docs"
    directory = "sweepai/core"
    directory = "sweepai/agents"
    # summarize_frameworks(cloned_repo)

    snippets, lexical_index = prepare_lexical_search_index(
        cloned_repo.repo_dir,
        SweepConfig(),
    )

    print(
        recursively_summarize_directory(
            cloned_repo,
        )
    )



from sweepai.core.entities import Message
from sweepai.core.llm.chat import LATEST_CLAUDE_SONNET_MODEL, Thread, continuous_llm_calls
from sweepai.utils.cache import create_cache
from sweepai.utils.format_utils import Prompt
from sweepai.utils.str_utils import safe_extract_xml_tag, tail

cache = create_cache()

ci_cleanup_system_prompt = (
    """You are an expert Devops engineer that extracts lines from failing Continuous Integration (CI) systems."""
)

ci_cleanup_user_prompt = Prompt(
    """# Task:
Your job is to extract ALL logs from the CI error logs that contain relevant information to fixing the issue.

{{ pulls_message }}


<error_logs>
{{ error_logs }}
</error_logs>

# Instructions

Copy the all of the important lines from this CI/CD run. Don't leave out any information that helps solve the failing CI/CD. 
If there is a certain type of error such as a missing dependency, include all specific instances of (file path x error) so all of the instances can be fixed.
If the exact same error occurs multiple times in one file, only include the first instance of the error. But if the error occurs in multiple files, include all files where the error occurs.

# Format:
Respond in this format:

<log_extraction_analysis>
1. Identify the main error(s).
2. Identify any patterns or recurring problems observed.
3. Then, identify all irrelevant information.
4. Indicate if any debug log statements were added to the error logs.
</log_extraction_analysis>

<most_important_lines>
[First, copy the CI command that failed.]

[Second, copy the important logs here.]
</most_important_lines>"""
)


def clean_ci_logs(
    username: str,
    error_logs: str,
    pulls_message: str,
    max_lines: int = 1000,
) -> str:
    # Keep only the last 1000 lines of error_logs
    error_logs = tail(error_logs, max_lines)
    thread = Thread(
        messages=[
            Message(role="system", content=ci_cleanup_system_prompt),
        ]
    )
    content = ci_cleanup_user_prompt.render(
        error_logs=error_logs,
        pulls_message=pulls_message,
    )
    response = continuous_llm_calls(
        username=username,
        thread=thread,
        content=content,
        stop_sequences=["</most_important_lines>"],
        model=LATEST_CLAUDE_SONNET_MODEL,
    )
    return safe_extract_xml_tag(response, "most_important_lines")


if __name__ == "__main__":
    # Testing the prompt
    prompt = ci_cleanup_user_prompt.render(error_logs="test", pulls_message="pull")
    assert prompt.startswith("# Task:")
    assert "<error_logs>" in prompt


import os
import traceback
from time import sleep, time

import backoff
from git import GitCommandError
from github.PullRequest import PullRequest
from github.Repository import Repository
from loguru import logger

from sweepai.config.server import GITHUB_LABEL_NAME
from sweepai.core.github_utils import ClonedRepo, get_github_client, refresh_token
from sweepai.dataclasses.codereview import GroupedFilesForReview, PRReviewCommentThread
from sweepai.o11y.event_logger import posthog
from sweepai.o11y.posthog_trace import posthog_trace
from sweepai.review.review_utils import (
    cluster_patches,
    decompose_code_review_by_group,
    format_all_pr_changes_by_groups,
    format_comment_threads,
    format_pr_info,
    get_all_comments_for_review,
    get_pr_changes,
    get_pr_summary_from_patches,
    group_vote_review_pr,
)
from sweepai.utils.ticket_rendering_utils import create_update_review_pr_comment
from sweepai.web.validate_license import validate_license


@backoff.on_exception(
    backoff.expo,
    Exception,
    max_tries=2,
)
@posthog_trace
def review_pr(
    username: str,
    pr: PullRequest,
    repository: Repository,
    installation_id: int,
    tracking_id: str | None = None,
    metadata: dict = {},
):
    if not os.environ.get("CLI"):
        assert (
            validate_license()
        ), "License key is invalid or expired. Please contact us at team@sweep.dev to upgrade to an enterprise license."
    with logger.contextualize(
        tracking_id=tracking_id,
    ):
        review_pr_start_time = time()
        metadata = {
            "pr_url": pr.html_url,
            "repo_full_name": repository.full_name,
            "repo_description": repository.description,
            "username": username,
            "installation_id": installation_id,
            "function": "review_pr",
            "tracking_id": tracking_id,
        }

        try:
            # check if the pr has been merged or not
            pr_labelled = GITHUB_LABEL_NAME in [label.name.lower() for label in pr.get_labels()]
            if pr.state == "closed" and not pr_labelled:
                posthog.capture(
                    username,
                    "pr_review pr_closed",
                    properties={
                        **metadata,
                        "duration": round(time() - review_pr_start_time),
                    },
                )
                return {"success": False, "reason": "PR is closed"}
            user_token, _ = get_github_client(installation_id)
            error: Exception = None

            try:
                # check if this is a pr from a forked repo
                if pr.head.repo.full_name != pr.base.repo.full_name:
                    error = Exception("Sweep does not support reviewing PRs from forked repositories.")
                    raise error
                sleep(10)  # sleep for 10 seconds to prevent race conditions with github uploading remote branch
                try:
                    cloned_repo: ClonedRepo = ClonedRepo(
                        repository.full_name,
                        installation_id=installation_id,
                        token=user_token,
                        github_repo=repository,
                    )
                    cloned_repo.safely_checkout_to_non_default_branch(pr.head.sha)
                except GitCommandError as e:
                    raise e
                except Exception:
                    error = Exception(
                        f"Failed to clone repository: {repository.full_name}. This may be because Sweep does not have the necessary permissions to access your repository."
                    )
                    raise error
            except GitCommandError as e:
                raise e
            except Exception as e:
                logger.error(f"Failure cloning repo in review_pr: {e}")
                error = e
            # try and update the user to let them know why we can not review the pr.
            # if the error is due to credential issues, this will probably not work
            if error:
                _comment_id = create_update_review_pr_comment(
                    username,
                    pr,
                    {},
                    error_message=str(error),
                )
                return {"success": False, "reason": str(error)}
            pr_issue = repository.get_issue(number=pr.number)
            reaction_eyes = pr_issue.create_reaction("eyes")
            # get all comments on the pr
            comment_threads: dict[str, list[PRReviewCommentThread]] = get_all_comments_for_review(
                repository.full_name, pr, installation_id
            )
            formatted_comment_threads: dict[str, str] = format_comment_threads(comment_threads)
            # handle creating comments on the pr to tell the user we are going to begin reviewing the pr
            pr_changes, dropped_files, unsuitable_files = get_pr_changes(pr, cloned_repo)
            # -1 group key means review those seperately
            grouped_files: dict[str, list[str]] = cluster_patches(pr_changes)
            # build another dict so that all files are in their own group
            single_files = {file_name: [file_name] for file_name in pr_changes.keys()}
            # render all groups of files
            formatted_pr_changes_by_group: dict[str, GroupedFilesForReview] = format_all_pr_changes_by_groups(
                grouped_files, pr_changes
            )
            # also render them individually
            formatted_pr_changes_by_file: dict[str, GroupedFilesForReview] = format_all_pr_changes_by_groups(
                single_files, pr_changes
            )
            # formatted_pr_changes_by_file = format_pr_changes_by_file(pr_changes)
            pull_request_info = format_pr_info(pr)
            # only get sweep to generate a summary if the pr doesnt have a description
            pull_request_summary = ""
            pull_request_summary = get_pr_summary_from_patches(pr_changes, pull_request_info=pull_request_info)

            # get initial code review by group vote
            code_review_by_group = group_vote_review_pr(
                username,
                pr_changes,
                formatted_pr_changes_by_group,
                formatted_pr_changes_by_file,
                cloned_repo,
                pull_request_info,
                formatted_comment_threads,
                pull_request_summary,
                multiprocess=True,
            )
            # convert code_review_by_group to be by file for easier rendering
            code_review_by_file = decompose_code_review_by_group(code_review_by_group)
            # do more specific checks now, i.e. duplicated util functions
            # after 50 minutes have passed refresh token to re get pr
            if time() - review_pr_start_time > 50 * 60:
                _, _, repository = refresh_token(repository.full_name, installation_id)
                pr = repository.get_pull(pr.number)
            # dont provide our pr summary if a pr description is provided
            if "pr_description" in pull_request_info:
                pull_request_summary = ""
            _comment_id = create_update_review_pr_comment(
                username,
                pr,
                formatted_comment_threads,
                code_review_by_file=code_review_by_file,
                pull_request_summary=pull_request_summary,
                dropped_files=dropped_files,
                unsuitable_files=unsuitable_files,
            )
            pr_issue.delete_reaction(reaction_eyes.id)
        except Exception as e:
            posthog.capture(
                username,
                "review_pr failed",
                properties={
                    **metadata,
                    "error": str(e),
                    "trace": traceback.format_exc(),
                    "duration": round(time() - review_pr_start_time),
                },
            )
            pr_issue.delete_reaction(reaction_eyes.id)
            raise e
        logger.info("review_pr success in " + str(round(time() - review_pr_start_time)))
        return {"success": True}



import re
from collections import defaultdict

from loguru import logger

from sweepai.core.github_utils import ClonedRepo

gha_context_cleanup_system_prompt = """You are a skilled programmer. You will be given a set of failing GitHub Action logs. Your sole task is to extract and return only the crucial parts that will help a developer resolve the issues. 
Eliminate any unnecessary context and return verbatim the useful lines of logs."""

gha_generate_query_system_prompt = """You are a skilled programmer. You will be given a set of failing GitHub Action logs. 
Your sole task is to create a query based on the failing Github Action logs that will be used to vector search a code base in order to fetch relevant code snippets.
The retrived code snippets will then be used as context to help the next developer resolve the failing Github Action logs."""

gha_context_cleanup_user_prompt = """
# The failing Github Action logs are given below

{github_actions_logs}

ONLY RETURN THE USEFUL PARTS OF THE LOGS THAT WILL HELP A DEVELOPER RESOLVE THE ISSUES. NOTHING ELSE.
"""

gha_generate_query_user_prompt = """
# The failing Github Action logs are given below

{github_actions_logs}

# The initial task that resulted in these failing Github Action logs is given below

{issue_description}

Return your query below in the following xml tags:
<query>
{{generated query goes here. Remember that it will be used in a vector search so tailor your query with that in mind. There will not be multiple searches, this one search needs to get all the revelant code snippets for all issues}}
</query>
"""

SLEEP_DURATION_SECONDS = 5


def get_error_locations_from_error_logs(error_logs: str, cloned_repo: ClonedRepo):
    pattern = re.compile(
        r"^(?P<file_path>.*?)[^a-zA-Z\d]+(?P<line_num>\d+)[^a-zA-Z\d]+(?P<col_num>\d+)[^a-zA-Z\d]+(?P<error_message>.*?)$",
        re.MULTILINE,
    )
    matches = list(pattern.finditer(error_logs))

    matched_files = []
    errors = defaultdict(dict)
    error_message = ""
    try:
        file_paths = cloned_repo.get_file_list()
        for match in matches:
            potential_file_path = match.group("file_path")
            line_number = match.group("line_num")
            current_error_message = match.group("error_message")
            if not any(file_path in potential_file_path for file_path in file_paths):
                # include raw message if we cannot find the file
                error_message += f"{match.group(0)}\n"
                continue
            actual_file_path = [file_path for file_path in file_paths if file_path in potential_file_path][0]
            matched_files.append(actual_file_path)

            errors[actual_file_path][int(line_number)] = current_error_message  # assume one error per line for now

        for file_path, errors_dict in errors.items():
            error_message += f"Here are the {len(errors_dict)} errors in {file_path}, each denotated by FIXME:\n```\n"
            file_contents = cloned_repo.get_file_contents(file_path)
            lines = file_contents.splitlines()
            erroring_lines = set()
            surrounding_lines = 5
            for line_number in errors_dict.keys():
                erroring_lines |= set(range(line_number - surrounding_lines, line_number + surrounding_lines))
            erroring_lines &= set(range(len(lines)))
            width = len(str(len(lines)))
            for i in sorted(list(erroring_lines)):
                if i not in erroring_lines:
                    error_message += "...\n"
                error_message += str(i + 1).ljust(width) + f" | {lines[i + 1]}"
                if i + 1 in errors_dict:
                    error_message += f"     FIXME {errors_dict[i + 1].strip()}"
                error_message += "\n"
            if len(lines) not in erroring_lines:
                error_message += "...\n"
            error_message += "```\n"
        deduped_matched_files = []
        for file_path in matched_files:
            if file_path not in deduped_matched_files:
                deduped_matched_files.append(file_path)
        if (
            not error_message
        ):  # monkey patch because this can fail and return empty error_message, which is not what we want
            return error_logs, []
        return error_message, deduped_matched_files
    except Exception as e:
        logger.error(f"Error in getting error locations: {e}")
        return error_logs, []
